import numpy as np
import multiprocessing as mp

from .core import (
    PartitionData,
    PartitionData_cp,
    RePartitionData,
    PrimitiveElasticGraphEmbedment,
    PrimitiveElasticGraphEmbedment_v2,
    PrimitiveElasticGraphEmbedment_v3,
    PrimitiveElasticGraphEmbedment_cp,
    PrimitiveElasticGraphEmbedment_cp_v2,
    PrimitiveElasticGraphBarycentricEmbedment,
    DecodeElasticMatrix2,
)
from .supervised import (
    gen_pseudotime_augmented_graph,
    gen_pseudotime_augmented_graph_by_path,
    old_gen_pseudotime_augmented_graph_by_path,
)
from .._EMAdjustment import AdjustByConstant


def proxy(Dict):
    return PrimitiveElasticGraphEmbedment(**Dict)


def proxy_cp(Dict):
    return PrimitiveElasticGraphEmbedment_cp(**Dict)


# Grammar function wrapper ------------------------------------------------


def GraphGrammarOperation(
    X,
    NodePositions,
    ElasticMatrix,
    AdjustVect,
    Type,
    partition,
    FixNodesAtPoints=[],
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
    PointWeights=None,
):

    if Type == "addnode2node":
        return AddNode2Node(
            X,
            NodePositions,
            ElasticMatrix,
            partition,
            AdjustVect,
            FixNodesAtPoints,
            MaxNumberOfGraphCandidates=MaxNumberOfGraphCandidatesDict[
                "AddNode2Node"
            ],
            PointWeights=PointWeights,
        )
    elif Type == "addnode2node_1":
        return AddNode2Node(
            X,
            NodePositions,
            ElasticMatrix,
            partition,
            AdjustVect,
            FixNodesAtPoints,
            Max_K=1,
            PointWeights=PointWeights,
        )
    elif Type == "addnode2node_2":
        return AddNode2Node(
            X,
            NodePositions,
            ElasticMatrix,
            partition,
            AdjustVect,
            FixNodesAtPoints,
            Max_K=2,
            PointWeights=PointWeights,
        )
    elif Type == "removenode":
        return RemoveNode(
            NodePositions, ElasticMatrix, AdjustVect, FixNodesAtPoints
        )
    elif Type == "bisectedge":
        return BisectEdge(
            NodePositions,
            ElasticMatrix,
            AdjustVect,
            MaxNumberOfGraphCandidates=MaxNumberOfGraphCandidatesDict[
                "BisectEdge"
            ],
        )
    elif Type == "bisectedge_3":
        return BisectEdge(NodePositions, ElasticMatrix, AdjustVect, Min_K=3)
    elif Type == "shrinkedge":
        return ShrinkEdge(
            NodePositions,
            ElasticMatrix,
            AdjustVect,
            MaxNumberOfGraphCandidates=MaxNumberOfGraphCandidatesDict[
                "ShrinkEdge"
            ],
        )
    elif Type == "shrinkedge_3":
        return ShrinkEdge(NodePositions, ElasticMatrix, AdjustVect, Min_K=3)
    else:
        raise ValueError("Operation " + Type + " is not defined")


# Grammar functions ------------------------------------------------


def RemoveNode(NodePositions, ElasticMatrix, AdjustVect, FixNodesAtPoints):
    """
    ##  This grammar operation removes a leaf node (connectivity==1)
    """
    Lambda = ElasticMatrix.copy()
    np.fill_diagonal(Lambda, 0)
    Connectivities = (Lambda > 0).sum(axis=0)

    # Define sizes
    nNodes = ElasticMatrix.shape[0]
    if FixNodesAtPoints != []:
        nGraphs = (Connectivities == 1).sum() - len(FixNodesAtPoints)
        start = len(FixNodesAtPoints)
    else:
        nGraphs = (Connectivities == 1).sum()
        start = 0

    # Preallocate arrays
    NodePositionsArray = [
        np.zeros((nNodes - 1, NodePositions.shape[1])) for i in range(nGraphs)
    ]
    ElasticMatrices = [
        np.zeros((nNodes - 1, nNodes - 1)) for i in range(nGraphs)
    ]
    NodeIndicesArray = np.zeros((nNodes - 1, nGraphs), dtype=int)
    AdjustVectArray = [[] for i in range(nGraphs)]

    k = 0
    for i in range(start, Connectivities.shape[0]):
        if Connectivities[i] == 1:
            # if terminal node remove it
            newInds = np.concatenate(
                (
                    np.arange(0, i, dtype=int),
                    np.arange(i + 1, nNodes, dtype=int),
                )
            )
            AdjustVectArray[k] = [AdjustVect[j] for j in newInds]
            NodePositionsArray[k] = NodePositions[newInds, :]
            tmp = np.repeat(False, nNodes)
            tmp[newInds] = True
            tmp2 = ElasticMatrix[tmp, :]
            ElasticMatrices[k] = tmp2[:, tmp]
            NodeIndicesArray[:, k] = newInds
            k += 1
    return (
        NodePositionsArray,
        ElasticMatrices,
        AdjustVectArray,
        NodeIndicesArray,
    )


def BisectEdge(
    NodePositions,
    ElasticMatrix,
    AdjustVect,
    Min_K=1,
    MaxNumberOfGraphCandidates=float("inf"),
):
    """
    # % This grammar operation inserts a node inside the middle of each edge
    # % The elasticity of the edges do not change
    # % The elasticity of the newborn star is chosen as
    # % mean over the neighbour stars if the edge connects two star centers
    # % or
    # % the one of the single neigbour star if this is a dangling edge
    # % or
    # % if one starts from a single edge, the star elasticities should be on
    # % one of two elements in the diagonal of the ElasticMatrix
    """
    # Decompose Elastic Matrix: Mus
    Mus = ElasticMatrix.diagonal()
    # Get list of edges
    Edges, _, _ = DecodeElasticMatrix2(ElasticMatrix)

    # Define some constants
    nNodes = NodePositions.shape[0]
    if Min_K > 1:
        Degree = np.bincount(Edges.flatten())
        EdgDegree = np.max(Degree[Edges], axis=1)
        nGraphs = np.where(EdgDegree >= Min_K)[0]
    else:
        nGraphs = np.array(range(Edges.shape[0]))

    # In case we have limits on the number of candidates
    if MaxNumberOfGraphCandidates < len(nGraphs):
        edge_lengths = np.sum(
            (NodePositions[Edges[:, 0], :] - NodePositions[Edges[:, 1], :]).T
            ** 2,
            axis=0,
        )
        inds = np.argsort(edge_lengths)[::-1]
        if Min_K > 1:
            nGraphs = inds[np.isin(inds, nGraphs)][:MaxNumberOfGraphCandidates]
            Edges = Edges[nGraphs]
        else:
            nGraphs = nGraphs[:MaxNumberOfGraphCandidates]
            Edges = Edges[inds[nGraphs]]

    # Create prototypes for new NodePositions, ElasticMatrix and inds
    npProt = np.vstack((NodePositions, np.zeros((1, NodePositions.shape[1]))))
    emProt = np.vstack(
        (
            np.hstack((ElasticMatrix, np.zeros((nNodes, 1)))),
            np.zeros((1, nNodes + 1)),
        )
    )
    niProt = np.arange(nNodes + 1)
    # niProt[nNodes] = 0

    # Allocate arrays and put prototypes in place
    NodePositionsArray = [npProt.copy() for i in range(len(nGraphs))]
    ElasticMatrices = [emProt.copy() for i in range(len(nGraphs))]
    NodeIndicesArray = np.repeat(niProt[:, np.newaxis], len(nGraphs), axis=1)
    AdjustVectArray = [AdjustVect + [False] for i in range(len(nGraphs))]

    for j, i in enumerate(nGraphs):
        NewNodePosition = (
            NodePositions[
                Edges[i, 0],
            ]
            + NodePositions[
                Edges[i, 1],
            ]
        ) / 2

        # Fill NodePosition
        NodePositionsArray[j][nNodes, :] = NewNodePosition
        # correct elastic matrix
        Lambda = ElasticMatrix[Edges[i, 0], Edges[i, 1]]
        # remove edge
        ElasticMatrices[j][Edges[i, 0], Edges[i, 1]] = 0
        ElasticMatrices[j][Edges[i, 1], Edges[i, 0]] = 0
        # add 2 edges
        ElasticMatrices[j][Edges[i, 0], nNodes] = Lambda
        ElasticMatrices[j][nNodes, Edges[i, 0]] = Lambda
        ElasticMatrices[j][nNodes, Edges[i, 1]] = Lambda
        ElasticMatrices[j][Edges[i, 1], nNodes] = Lambda
        # Define mus of edges
        mu1 = Mus[Edges[i, 0]]
        mu2 = Mus[Edges[i, 1]]
        if mu1 > 0 and mu2 > 0:
            ElasticMatrices[j][nNodes, nNodes] = (mu1 + mu2) / 2
        else:
            ElasticMatrices[j][nNodes, nNodes] = max(mu1, mu2)

    return (
        NodePositionsArray,
        ElasticMatrices,
        AdjustVectArray,
        NodeIndicesArray,
    )


def AddNode2Node(
    X,
    NodePositions,
    ElasticMatrix,
    partition,
    AdjustVect,
    FixNodesAtPoints,
    Max_K=float("inf"),
    MaxNumberOfGraphCandidates=float("inf"),
    PointWeights=None,
):
    """
    Adds a node to each graph node

    This grammar operation adds a node to each graph node. The position of the node
    is chosen as a linear extrapolation for a leaf node (in this case the elasticity of
    a newborn star is chosed as in BisectEdge operation), or as the data point giving
    the minimum local MSE for a star (without any optimization).

    X
    NodePositions
    ElasticMatrix
    @return


    @details



    @examples
    """
    nNodes = NodePositions.shape[0]
    Mus = ElasticMatrix.diagonal()
    Lambda = ElasticMatrix.copy()
    np.fill_diagonal(Lambda, 0)
    indL = Lambda > 0
    Connectivities = indL.sum(axis=0)
    # add pointweights here if added
    if PointWeights is not None:
        assoc = np.bincount(
            partition[partition > -1].ravel(),
            weights=PointWeights.ravel(),
            minlength=nNodes,
        )
    else:
        assoc = np.bincount(
            partition[partition > -1].ravel(), minlength=nNodes
        )

    # Create prototypes for new NodePositions, ElasticMatrix and inds
    npProt = np.vstack((NodePositions, np.zeros((1, NodePositions.shape[1]))))
    emProt = np.vstack(
        (np.hstack((Lambda, np.zeros((nNodes, 1)))), np.zeros((1, nNodes + 1)))
    )
    niProt = np.arange(nNodes + 1, dtype=int)
    #     niProt[nNodes] = 0

    MuProt = np.zeros(nNodes + 1)
    MuProt[:-1] = Mus

    if not np.isinf(Max_K):
        Degree = np.sum(ElasticMatrix > 0, axis=1)
        Degree[Degree > 1] = Degree[Degree > 1] - 1

        if np.sum(Degree <= Max_K) > 1:
            idx_nodes = np.where(Degree <= Max_K)[0]
        else:
            raise ValueError(
                "AddNode2Node impossible with the current parameters!"
            )
    else:
        idx_nodes = np.array(range(nNodes))

    # In case we have limits on the number of candidates
    if MaxNumberOfGraphCandidates < len(idx_nodes) and np.isinf(Max_K):
        idx_nodes = np.argsort(assoc)[::-1][:MaxNumberOfGraphCandidates]

    elif MaxNumberOfGraphCandidates < len(idx_nodes) and not (np.isinf(Max_K)):
        nGraphs = [i for i in np.argsort(assoc)[::-1] if i in idx_nodes]
        idx_nodes = np.array(nGraphs)[:MaxNumberOfGraphCandidates]

    # In case we have fixed nodes
    if FixNodesAtPoints != []:
        idx_nodes = idx_nodes[
            ~np.isin(idx_nodes, np.arange(len(FixNodesAtPoints)))
        ]

    # Put prototypes to corresponding places
    NodePositionsArray = [npProt.copy() for i in range(len(idx_nodes))]
    ElasticMatrices = [emProt.copy() for i in range(len(idx_nodes))]
    NodeIndicesArray = np.repeat(niProt[:, np.newaxis], len(idx_nodes), axis=1)
    AdjustVectArray = [AdjustVect + [False] for i in range(len(idx_nodes))]

    for j, i in enumerate(idx_nodes):
        MuProt[-1] = 0
        # Compute mean edge elasticity for edges with node i
        meanL = Lambda[
            i,
            indL[
                i,
            ],
        ].mean(axis=0)
        # Add edge to elasticity matrix
        ElasticMatrices[j][nNodes, i] = ElasticMatrices[j][i, nNodes] = meanL

        if Connectivities[i] == 1:
            # Add node to terminal node
            ineighbour = np.nonzero(
                indL[
                    i,
                ]
            )[0]
            # Calculate new node position
            NewNodePosition = (
                2
                * NodePositions[
                    i,
                ]
                - NodePositions[
                    ineighbour,
                ]
            )
            # Complete Elasticity Matrix
            MuProt[i] = Mus[ineighbour]
        else:
            # Add node to a star
            # if 0 data points associated with this star
            if assoc[i] == 0:
                # then select mean of all leaves as new position
                NewNodePosition = NodePositions[indL[:, i]].mean(axis=0)

            else:
                # Otherwise take the mean of the points associated with the
                # central node
                NewNodePosition = X[(partition == i).ravel()].mean(axis=0)
        # fill node position
        NodePositionsArray[j][nNodes, :] = NewNodePosition
        np.fill_diagonal(ElasticMatrices[j], MuProt)

    return (
        NodePositionsArray,
        ElasticMatrices,
        AdjustVectArray,
        NodeIndicesArray,
    )


def ShrinkEdge(
    NodePositions,
    ElasticMatrix,
    AdjustVect,
    Min_K=1,
    MaxNumberOfGraphCandidates=float("inf"),
):
    """
    # %
    # % This grammar operation removes an edge from the graph
    # % If this is an edge connecting a leaf node then it is equivalent to
    # % RemoveNode. So we remove only internal edges.
    # % If this is an edge connecting two stars then their leaves are merged,
    # % and the star is placed in the middle of the shrinked edge.
    # % The elasticity of the new formed star is the average of two star
    # % elasticities.
    # %
    """
    ## Shrink edge
    Mus = ElasticMatrix.diagonal()
    Lambda = ElasticMatrix.copy()
    np.fill_diagonal(Lambda, 0)
    Connectivities = (Lambda > 0).sum(axis=0)
    # get list of edges
    start, stop = np.triu(ElasticMatrix, 1).nonzero()
    # Edges = np.concatenate((start[None], stop[None]))
    # define size
    nNodes = NodePositions.shape[0]
    # identify edges with minimal connectivity > 1
    Degree = np.hstack(
        (Connectivities[start[None].T], Connectivities[stop[None].T])
    )

    ind_sup1 = np.min(Degree, axis=1) > 1
    ind_min_K = np.max(Degree, axis=1) >= Min_K
    ind = np.where(ind_sup1 & ind_min_K)[0]

    # calculate nb of graphs
    nGraphs = len(ind)

    # In case we have limits on the number of candidates
    if MaxNumberOfGraphCandidates < nGraphs:
        edge_lengths = np.sum(
            (NodePositions[start, :] - NodePositions[stop, :]).T ** 2,
            axis=0,
        )
        nGraphs = MaxNumberOfGraphCandidates
        ind_l = np.argsort(edge_lengths)
        ind_l = ind_l[np.isin(ind_l, ind)][:nGraphs]
        start, stop = start[ind_l], stop[ind_l]
    else:
        start, stop = start[ind], stop[ind]

    # preallocate array
    NodePositionsArray = [
        np.zeros((nNodes - 1, NodePositions.shape[1])) for i in range(nGraphs)
    ]
    ElasticMatrices = [
        np.zeros((nNodes - 1, nNodes - 1)) for i in range(nGraphs)
    ]
    NodeIndicesArray = np.zeros((nNodes - 1, nGraphs), dtype=int)
    AdjustVectArray = [[] for i in range(nGraphs)]

    for i in range(nGraphs):
        # create copy of elastic matrix
        em = ElasticMatrix.copy()
        # Reattaches all edges connected with stop[i] to start[i]
        # and make a new star with an elasticity average of two merged stars
        em[start[i],] = np.maximum(
            Lambda[
                start[i],
            ],
            Lambda[
                stop[i],
            ],
        )
        em[:, start[i]] = np.maximum(Lambda[:, start[i]], Lambda[:, stop[i]])
        em[start[i], start[i]] = (Mus[start[i]] + Mus[stop[i]]) / 2
        #         em[start[i], start[i]] = Mus[start[i]] + Mus[stop[i]] / 2  #### R bug ????
        # Create copy of node positions
        nodep = NodePositions.copy()
        # modify node start[i]
        nodep[start[i], :] = (nodep[start[i], :] + nodep[stop[i], :]) / 2
        # Form index for retained nodes and extract corresponding part of
        # node positions and elastic matrix
        newInds = np.concatenate(
            (np.arange(0, stop[i]), np.arange(stop[i] + 1, nNodes))
        )

        AdjustVectArray[i] = [AdjustVect[j] for j in newInds]

        NodePositionsArray[i] = nodep[newInds, :]
        ElasticMatrices[i] = em.take(newInds, axis=0).take(newInds, axis=1)
        NodeIndicesArray[:, i] = newInds

    return (
        NodePositionsArray,
        ElasticMatrices,
        AdjustVectArray,
        NodeIndicesArray,
    )


def ApplyOptimalGraphGrammarOperation(
    X,
    NodePositions,
    ElasticMatrix,
    opTypes,
    AdjustVect=None,
    SquaredX=None,
    verbose=False,
    MaxBlockSize=100000000,
    MaxNumberOfIterations=100,
    eps=0.01,
    TrimmingRadius=float("inf"),
    Mode=1,
    FinalEnergy="Base",
    alpha=1,
    beta=1,
    EmbPointProb=1,
    PointWeights=None,
    AvoidSolitary=False,
    AdjustElasticMatrix=None,
    DisplayWarnings=True,
    n_cores=1,
    MinParOp=20,
    multiproc_shared_variables=None,
    Xcp=None,
    SquaredXcp=None,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    # Multiple grammar application --------------------------------------------
    Application of the grammar operation. This in an internal function that should not be used in by the end-user

    X numerical 2D matrix, the n-by-m matrix with the position of n m-dimensional points
    NodePositions numerical 2D matrix, the k-by-m matrix with the position of k m-dimensional points
    ElasticMatrix numerical 2D matrix, the k-by-k elastic matrix
    operationtypes string vector containing the operation to use
    SquaredX rowSums(X^2), if NULL it will be computed
    verbose boolean. Should addition information be displayed
    n.cores integer. How many cores to use. If EnvCl is not NULL, that cliuster setup will be used,
    otherwise a SOCK cluster willbe used
    EnvCl a cluster structure returned, e.g., by makeCluster.
    If a cluster structure is used, all the nodes must be able to access all the variable needed by PrimitiveElasticGraphEmbedment
    MaxNumberOfIterations is an integer number indicating the maximum number of iterations for the EM algorithm
    TrimmingRadius is a real value indicating the trimming radius, a parameter required for robust principal graphs
    (see https://github.com/auranic/Elastic-principal-graphs/wiki/Robust-principal-graphs)
    eps a real number indicating the minimal relative change in the nodenpositions
    to be considered the graph embedded (convergence criteria)
    Mode integer, the energy mode. It can be 1 (difference is computed using the position of the nodes) and
    2 (difference is computed using the changes in elestic energy of the configuraztions)
    FinalEnergy string indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha positive numeric, the value of the alpha parameter of the penalized elastic energy
    beta positive numeric, the value of the beta parameter of the penalized elastic energy
    gamma
    FastSolve boolean, should FastSolve be used when fitting the points to the data?
    AvoidSolitary boolean, should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    EmbPointProb numeric between 0 and 1. If less than 1 point will be sampled at each iteration. Prob indicate the probability of
    using each points. This is an *experimental* feature, which may helps speeding up the computation if a large number of points is present.
    AdjustVect
    AdjustElasticMatrix
    ...
    MinParOp integer, the minimum number of operations to use parallel computation

    @return

    @examples
    """

    NodePositionsArrayAll = []
    ElasticMatricesAll = []
    AdjustVectAll = []
    NodeIndicesArrayAll = []
    opTypesAll = []
    if Xcp is None:
        partition, _ = PartitionData(
            X, NodePositions, MaxBlockSize, SquaredX, TrimmingRadius
        )
    else:
        partition, _ = PartitionData_cp(
            Xcp, NodePositions, MaxBlockSize, SquaredXcp, TrimmingRadius
        )

    for i in range(len(opTypes)):
        if verbose:
            print(" Operation type : ", opTypes[i])
        (
            NodePositionsArray,
            ElasticMatrices,
            AdjustVectArray,
            NodeIndicesArray,
        ) = GraphGrammarOperation(
            X,
            NodePositions,
            ElasticMatrix,
            AdjustVect,
            opTypes[i],
            partition,
            FixNodesAtPoints,
            MaxNumberOfGraphCandidatesDict,
        )

        #         NodePositionsArrayAll = np.concatenate((NodePositionsArrayAll,
        #                                                NodePositionsArray), axis=2)
        #         ElasticMatricesAll = np.concatenate((ElasticMatricesAll,
        #                                              ElasticMatrices), axis=2)
        #     NodeIndicesArrayAll = np.concatenate((NodeIndicesArrayAll,
        #                                           NodeIndicesArray), axis=1)
        NodePositionsArrayAll.extend(NodePositionsArray)
        ElasticMatricesAll.extend(ElasticMatrices)
        AdjustVectAll.extend(AdjustVectArray)
        NodeIndicesArrayAll.extend(NodeIndicesArray.T)
        opTypesAll.extend([opTypes[i]] * len(NodePositionsArray))

    if verbose:
        print("Optimizing graphs")

    Valid_configurations = range(len(NodePositionsArrayAll))

    if AvoidSolitary:
        Valid_configurations = []
        if Xcp is None:
            for i in range(len(NodePositionsArrayAll)):
                partition = PartitionData(
                    X=X,
                    MaxBlockSize=MaxBlockSize,
                    NodePositions=NodePositionsArrayAll[i],
                    SquaredX=SquaredX,
                    TrimmingRadius=TrimmingRadius,
                )[0]
                if all(
                    np.isin(
                        np.array(range(NodePositionsArrayAll[i].shape[0])),
                        partition[partition > -1],
                    )
                ):
                    Valid_configurations.append(i)

        else:
            for i in range(len(NodePositionsArrayAll)):
                partition = PartitionData_cp(
                    Xcp,
                    MaxBlockSize,
                    NodePositionsArrayAll[i],
                    SquaredXcp,
                    TrimmingRadius=TrimmingRadius,
                )[0]
                if all(
                    np.isin(
                        np.array(range(NodePositionsArrayAll[i].shape[0])),
                        partition[partition > -1],
                    )
                ):
                    Valid_configurations.append(i)

        if verbose:
            print(
                len(Valid_configurations),
                "configurations out of ",
                len(NodePositionsArrayAll),
                "used",
            )
        if Valid_configurations == []:
            return "failed operation"
    #     NodePositionArrayAll = NodePositionArrayAll[...,Valid_configurations]
    #     ElasticMatricesAll = ElasticMatricesAll[...,Valid_configurations]
    #     AdjustVectAll = AdjustVectAll[Valid_configurations]

    if AdjustElasticMatrix:
        for i in Valid_configurations:
            ElasticMatricesAll[i], AdjustVectAll[i] = AdjustByConstant(
                ElasticMatricesAll[i], AdjustVectAll[i]
            )

    if n_cores > 1 and len(Valid_configurations) // (MinParOp + 1) > 1:

        #             X_remote, X_shape, SquaredX_remote, SquaredX_shape = multiproc_shared_variables
        #             ####### Multiprocessing
        #             with Pool(processes=n_cores,
        #                       initializer=init_worker,
        #                       initargs=(X_remote,
        #                                 X_shape,
        #                                 SquaredX_remote,
        #                                 SquaredX_shape,
        #                                 NodePositionsArrayAll,
        #                                 ElasticMatricesAll,
        #                                 MaxNumberOfIterations,
        #                                 TrimmingRadius,
        #                                 eps,
        #                                 Mode,
        #                                 FinalEnergy,
        #                                 alpha,
        #                                 beta,
        #                                 EmbPointProb,
        #                                 DisplayWarnings,
        #                                 None,
        #                                 MaxBlockSize,
        #                                 False)) as pool:

        #                 results = pool.map(proxy_multiproc,Valid_configurations)

        with mp.Pool(n_cores) as pool:
            if Xcp is None:
                results = pool.map(
                    proxy,
                    [
                        dict(
                            X=X,
                            NodePositions=NodePositionsArrayAll[i],
                            ElasticMatrix=ElasticMatricesAll[i],
                            MaxNumberOfIterations=MaxNumberOfIterations,
                            eps=eps,
                            Mode=Mode,
                            FinalEnergy=FinalEnergy,
                            alpha=alpha,
                            beta=beta,
                            prob=EmbPointProb,
                            DisplayWarnings=DisplayWarnings,
                            PointWeights=PointWeights,
                            MaxBlockSize=MaxBlockSize,
                            verbose=False,
                            TrimmingRadius=TrimmingRadius,
                            SquaredX=SquaredX,
                        )
                        for i in Valid_configurations
                    ],
                )
            else:
                results = pool.map(
                    proxy_cp,
                    [
                        dict(
                            X=X,
                            NodePositions=NodePositionsArrayAll[i],
                            ElasticMatrix=ElasticMatricesAll[i],
                            MaxNumberOfIterations=MaxNumberOfIterations,
                            eps=eps,
                            Mode=Mode,
                            FinalEnergy=FinalEnergy,
                            alpha=alpha,
                            beta=beta,
                            prob=EmbPointProb,
                            DisplayWarnings=DisplayWarnings,
                            PointWeights=PointWeights,
                            MaxBlockSize=MaxBlockSize,
                            verbose=False,
                            TrimmingRadius=TrimmingRadius,
                            SquaredX=SquaredX,
                            Xcp=Xcp,
                            SquaredXcp=SquaredXcp,
                        )
                        for i in Valid_configurations
                    ],
                )

        list_energies = [r[1] for r in results]
        idx = list_energies.index(min(list_energies))
        NewNodePositions, minEnergy, partition, Dist, MSE, EP, RP = results[
            idx
        ]
        AdjustVect = AdjustVectAll[idx]
        NewElasticMatrix = ElasticMatricesAll[idx]

    else:
        minEnergy = np.inf
        StoreMeanPseudotime = None
        StoreMergedElasticMatrix = None
        StoreMergedNodePositions = None
        if Xcp is None:
            for i in Valid_configurations:
                # TODO add pointweights ?
                if pseudotime is not None and len(NodePositions) > 10:
                    (
                        MeanPseudotime,
                        MergedNodePositions,
                        MergedElasticMatrix,
                        MergedEdges,
                        nPseudoNodes,
                    ) = gen_pseudotime_augmented_graph_by_path(
                        X,
                        SquaredX,
                        NodePositionsArrayAll[i],
                        ElasticMatricesAll[i],
                        pseudotime,
                        root_node=0,
                        LinkMu=0,
                        LinkLambda=pseudotimeLambda,
                        PointWeights=PointWeights,
                        TrimmingRadius=TrimmingRadius,
                    )

                    FitNodePositions = MergedNodePositions
                    FitElasticMatrix = MergedElasticMatrix
                else:
                    FitNodePositions = NodePositionsArrayAll[i]
                    FitElasticMatrix = ElasticMatricesAll[i]
                    nPseudoNodes = 0
                (
                    nodep,
                    ElasticEnergy,
                    part,
                    dist,
                    mse,
                    ep,
                    rp,
                ) = PrimitiveElasticGraphEmbedment(
                    X,
                    FitNodePositions,
                    FitElasticMatrix,
                    MaxNumberOfIterations,
                    eps,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,
                    prob=EmbPointProb,
                    DisplayWarnings=DisplayWarnings,
                    PointWeights=PointWeights,
                    MaxBlockSize=MaxBlockSize,
                    verbose=False,
                    TrimmingRadius=TrimmingRadius,
                    SquaredX=SquaredX,
                    FixNodesAtPoints=[[] for i in range(nPseudoNodes)]
                    + FixNodesAtPoints,
                )

                if ElasticEnergy < minEnergy:
                    if pseudotime is not None and len(NodePositions) > 10:
                        StoreMeanPseudotime = MeanPseudotime
                        StoreMergedElasticMatrix = MergedElasticMatrix
                        StoreMergedNodePositions = nodep
                        NewNodePositions = nodep[nPseudoNodes:]
                    else:
                        StoreMeanPseudotime = None
                        StoreMergedElasticMatrix = None
                        StoreMergedNodePositions = None
                        NewNodePositions = nodep
                    NewElasticMatrix = ElasticMatricesAll[i]
                    partition = part
                    AdjustVect = AdjustVectAll[i]
                    minEnergy = ElasticEnergy
                    MSE = mse
                    EP = ep
                    RP = rp
                    Dist = dist

        else:
            for i in Valid_configurations:
                # TODO add pointweights ?
                if pseudotime is not None and len(NodePositions) > 10:
                    (
                        MeanPseudotime,
                        MergedNodePositions,
                        MergedElasticMatrix,
                        MergedEdges,
                        nPseudoNodes,
                    ) = gen_pseudotime_augmented_graph_by_path(
                        X,
                        SquaredX,
                        NodePositionsArrayAll[i],
                        ElasticMatricesAll[i],
                        pseudotime,
                        root_node=0,
                        LinkMu=0,
                        LinkLambda=pseudotimeLambda,
                        PointWeights=PointWeights,
                        TrimmingRadius=TrimmingRadius,
                    )

                    FitNodePositions = MergedNodePositions
                    FitElasticMatrix = MergedElasticMatrix
                else:
                    FitNodePositions = NodePositionsArrayAll[i]
                    FitElasticMatrix = ElasticMatricesAll[i]
                    nPseudoNodes = 0
                (
                    nodep,
                    ElasticEnergy,
                    part,
                    dist,
                    mse,
                    ep,
                    rp,
                ) = PrimitiveElasticGraphEmbedment_cp(
                    X,
                    FitNodePositions,
                    FitElasticMatrix,
                    MaxNumberOfIterations,
                    eps,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,
                    prob=EmbPointProb,
                    DisplayWarnings=DisplayWarnings,
                    PointWeights=PointWeights,
                    MaxBlockSize=MaxBlockSize,
                    verbose=False,
                    TrimmingRadius=TrimmingRadius,
                    SquaredX=SquaredX,
                    Xcp=Xcp,
                    SquaredXcp=SquaredXcp,
                    FixNodesAtPoints=[[] for i in range(nPseudoNodes)]
                    + FixNodesAtPoints,
                )

                if ElasticEnergy < minEnergy:
                    if pseudotime is not None and len(NodePositions) > 10:
                        StoreMeanPseudotime = MeanPseudotime
                        StoreMergedElasticMatrix = MergedElasticMatrix
                        StoreMergedNodePositions = nodep
                        NewNodePositions = nodep[nPseudoNodes:]
                    else:
                        StoreMeanPseudotime = None
                        StoreMergedElasticMatrix = None
                        StoreMergedNodePositions = None
                        NewNodePositions = nodep
                    NewElasticMatrix = ElasticMatricesAll[i]
                    partition = part
                    AdjustVect = AdjustVectAll[i]
                    minEnergy = ElasticEnergy
                    MSE = mse
                    EP = ep
                    RP = rp
                    Dist = dist

        return dict(
            StoreMeanPseudotime=StoreMeanPseudotime,
            StoreMergedElasticMatrix=StoreMergedElasticMatrix,
            StoreMergedNodePositions=StoreMergedNodePositions,
            NodePositions=NewNodePositions,
            ElasticMatrix=NewElasticMatrix,
            ElasticEnergy=minEnergy,
            MSE=MSE,
            EP=EP,
            RP=RP,
            AdjustVect=AdjustVect,
            Dist=Dist,
        )


def ApplyOptimalGraphGrammarOperation_v2(
    X,
    NodePositions,
    ElasticMatrix,
    opTypes,
    AdjustVect=None,
    SquaredX=None,
    verbose=False,
    MaxBlockSize=100000000,
    MaxNumberOfIterations=100,
    eps=0.01,
    TrimmingRadius=float("inf"),
    Mode=1,
    FinalEnergy="Base",
    alpha=1,
    beta=1,
    EmbPointProb=1,
    PointWeights=None,
    AvoidSolitary=False,
    AdjustElasticMatrix=None,
    DisplayWarnings=True,
    n_cores=1,
    MinParOp=20,
    multiproc_shared_variables=None,
    Xcp=None,
    SquaredXcp=None,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    # Multiple grammar application --------------------------------------------
    Application of the grammar operation. This in an internal function that should not be used in by the end-user

    X numerical 2D matrix, the n-by-m matrix with the position of n m-dimensional points
    NodePositions numerical 2D matrix, the k-by-m matrix with the position of k m-dimensional points
    ElasticMatrix numerical 2D matrix, the k-by-k elastic matrix
    operationtypes string vector containing the operation to use
    SquaredX rowSums(X^2), if NULL it will be computed
    verbose boolean. Should addition information be displayed
    n.cores integer. How many cores to use. If EnvCl is not NULL, that cliuster setup will be used,
    otherwise a SOCK cluster willbe used
    EnvCl a cluster structure returned, e.g., by makeCluster.
    If a cluster structure is used, all the nodes must be able to access all the variable needed by PrimitiveElasticGraphEmbedment
    MaxNumberOfIterations is an integer number indicating the maximum number of iterations for the EM algorithm
    TrimmingRadius is a real value indicating the trimming radius, a parameter required for robust principal graphs
    (see https://github.com/auranic/Elastic-principal-graphs/wiki/Robust-principal-graphs)
    eps a real number indicating the minimal relative change in the nodenpositions
    to be considered the graph embedded (convergence criteria)
    Mode integer, the energy mode. It can be 1 (difference is computed using the position of the nodes) and
    2 (difference is computed using the changes in elestic energy of the configuraztions)
    FinalEnergy string indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha positive numeric, the value of the alpha parameter of the penalized elastic energy
    beta positive numeric, the value of the beta parameter of the penalized elastic energy
    gamma
    FastSolve boolean, should FastSolve be used when fitting the points to the data?
    AvoidSolitary boolean, should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    EmbPointProb numeric between 0 and 1. If less than 1 point will be sampled at each iteration. Prob indicate the probability of
    using each points. This is an *experimental* feature, which may helps speeding up the computation if a large number of points is present.
    AdjustVect
    AdjustElasticMatrix
    ...
    MinParOp integer, the minimum number of operations to use parallel computation

    @return

    @examples
    """

    NodePositionsArrayAll = []
    ElasticMatricesAll = []
    AdjustVectAll = []
    NodeIndicesArrayAll = []
    opTypesAll = []
    if Xcp is None:
        partition, dists, precomp_d = PartitionData(
            X,
            NodePositions,
            MaxBlockSize,
            SquaredX,
            TrimmingRadius,
            precomp=True,
        )
    else:
        partition, dists, precomp_d = PartitionData_cp(
            Xcp,
            NodePositions,
            MaxBlockSize,
            SquaredXcp,
            TrimmingRadius,
            precomp=True,
        )

    for i in range(len(opTypes)):
        if verbose:
            print(" Operation type : ", opTypes[i])
        (
            NodePositionsArray,
            ElasticMatrices,
            AdjustVectArray,
            NodeIndicesArray,
        ) = GraphGrammarOperation(
            X,
            NodePositions,
            ElasticMatrix,
            AdjustVect,
            opTypes[i],
            partition,
            FixNodesAtPoints,
            MaxNumberOfGraphCandidatesDict,
            PointWeights=PointWeights,
        )

        NodePositionsArrayAll.extend(NodePositionsArray)
        ElasticMatricesAll.extend(ElasticMatrices)
        AdjustVectAll.extend(AdjustVectArray)
        NodeIndicesArrayAll.extend(NodeIndicesArray.T)
        opTypesAll.extend([opTypes[i]] * len(NodePositionsArray))

    if verbose:
        print("Optimizing graphs")

    Valid_configurations = range(len(NodePositionsArrayAll))

    if AvoidSolitary:
        Valid_configurations = []
        if Xcp is None:
            for i in range(len(NodePositionsArrayAll)):
                partition = PartitionData(
                    X=X,
                    MaxBlockSize=MaxBlockSize,
                    NodePositions=NodePositionsArrayAll[i],
                    SquaredX=SquaredX,
                    TrimmingRadius=TrimmingRadius,
                )[0]
                if all(
                    np.isin(
                        np.array(range(NodePositionsArrayAll[i].shape[0])),
                        partition[partition > -1],
                    )
                ):
                    Valid_configurations.append(i)

        else:
            for i in range(len(NodePositionsArrayAll)):
                partition = PartitionData_cp(
                    Xcp,
                    MaxBlockSize,
                    NodePositionsArrayAll[i],
                    SquaredXcp,
                    TrimmingRadius=TrimmingRadius,
                )[0]
                if all(
                    np.isin(
                        np.array(range(NodePositionsArrayAll[i].shape[0])),
                        partition[partition > -1],
                    )
                ):
                    Valid_configurations.append(i)

        if verbose:
            print(
                len(Valid_configurations),
                "configurations out of ",
                len(NodePositionsArrayAll),
                "used",
            )
        if Valid_configurations == []:
            return "failed operation"
    #     NodePositionArrayAll = NodePositionArrayAll[...,Valid_configurations]
    #     ElasticMatricesAll = ElasticMatricesAll[...,Valid_configurations]
    #     AdjustVectAll = AdjustVectAll[Valid_configurations]

    if AdjustElasticMatrix:
        for i in Valid_configurations:
            ElasticMatricesAll[i], AdjustVectAll[i] = AdjustByConstant(
                ElasticMatricesAll[i], AdjustVectAll[i]
            )

    if n_cores > 1 and len(Valid_configurations) // (MinParOp + 1) > 1:

        #             X_remote, X_shape, SquaredX_remote, SquaredX_shape = multiproc_shared_variables
        #             ####### Multiprocessing
        #             with Pool(processes=n_cores,
        #                       initializer=init_worker,
        #                       initargs=(X_remote,
        #                                 X_shape,
        #                                 SquaredX_remote,
        #                                 SquaredX_shape,
        #                                 NodePositionsArrayAll,
        #                                 ElasticMatricesAll,
        #                                 MaxNumberOfIterations,
        #                                 TrimmingRadius,
        #                                 eps,
        #                                 Mode,
        #                                 FinalEnergy,
        #                                 alpha,
        #                                 beta,
        #                                 EmbPointProb,
        #                                 DisplayWarnings,
        #                                 None,
        #                                 MaxBlockSize,
        #                                 False)) as pool:

        #                 results = pool.map(proxy_multiproc,Valid_configurations)

        with mp.Pool(n_cores) as pool:
            results = pool.map(
                proxy,
                [
                    dict(
                        X=X,
                        NodePositions=NodePositionsArrayAll[i],
                        ElasticMatrix=ElasticMatricesAll[i],
                        MaxNumberOfIterations=MaxNumberOfIterations,
                        eps=eps,
                        Mode=Mode,
                        FinalEnergy=FinalEnergy,
                        alpha=alpha,
                        beta=beta,
                        prob=EmbPointProb,
                        DisplayWarnings=DisplayWarnings,
                        PointWeights=PointWeights,
                        MaxBlockSize=MaxBlockSize,
                        verbose=False,
                        TrimmingRadius=TrimmingRadius,
                        SquaredX=SquaredX,
                    )
                    for i in Valid_configurations
                ],
            )

        list_energies = [r[1] for r in results]
        idx = list_energies.index(min(list_energies))
        NewNodePositions, minEnergy, partition, Dist, MSE, EP, RP = results[
            idx
        ]
        AdjustVect = AdjustVectAll[idx]
        NewElasticMatrix = ElasticMatricesAll[idx]

    else:
        minEnergy = np.inf
        StoreMergedElasticMatrix = None
        StoreMergedNodePositions = None
        cache_PseudotimeNodePositions = {}
        for i in Valid_configurations:
            newpartition, newdists, newprecomp_d = RePartitionData(
                X,
                NodePositions,
                NodePositionsArrayAll[i],
                NodeIndicesArrayAll[i],
                opTypesAll[i],
                precomp_d,
                SquaredX,
                TrimmingRadius=np.inf,
            )
            if pseudotime is not None and len(NodePositions) > 10:
                (
                    MergedNodePositions,
                    MergedElasticMatrix,
                    MergedEdges,
                    nPseudoNodes,
                    cache_PseudotimeNodePositions,
                ) = gen_pseudotime_augmented_graph_by_path(
                    X,
                    SquaredX,
                    NodePositionsArrayAll[i],
                    ElasticMatricesAll[i],
                    pseudotime,
                    root_node=0,
                    LinkMu=0,
                    LinkLambda=pseudotimeLambda,
                    PointWeights=PointWeights,
                    TrimmingRadius=TrimmingRadius,
                    partition=newpartition,
                    cache_PseudotimeNodePositions=cache_PseudotimeNodePositions,
                )

                # (
                #    MeanPseudotime,
                #    MergedNodePositions,
                #    MergedElasticMatrix,
                #    MergedEdges,
                #    nPseudoNodes,
                # ) = old_gen_pseudotime_augmented_graph_by_path(
                #    X,
                #    SquaredX,
                #    NodePositionsArrayAll[i],
                #    ElasticMatricesAll[i],
                #    pseudotime,
                #    root_node=0,
                #    LinkMu=0,
                #    LinkLambda=pseudotimeLambda,
                #    PointWeights=PointWeights,
                #    TrimmingRadius=TrimmingRadius,
                #    partition=newpartition,
                # )

                _split = int(len(MergedNodePositions) / 2)
                FitNodePositions = MergedNodePositions[_split:]
                PseudotimeNodePositions = MergedNodePositions[:_split]
                FitElasticMatrix = ElasticMatricesAll[i]

            else:
                FitNodePositions = NodePositionsArrayAll[i]
                FitElasticMatrix = ElasticMatricesAll[i]
                PseudotimeNodePositions = None

            if Xcp is None:
                (
                    nodep,
                    ElasticEnergy,
                    part,
                    dist,
                    mse,
                    ep,
                    rp,
                ) = PrimitiveElasticGraphEmbedment_v2(
                    X,
                    FitNodePositions,
                    FitElasticMatrix,
                    MaxNumberOfIterations,
                    eps,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,
                    prob=EmbPointProb,
                    DisplayWarnings=DisplayWarnings,
                    PointWeights=PointWeights,
                    MaxBlockSize=MaxBlockSize,
                    verbose=False,
                    TrimmingRadius=TrimmingRadius,
                    SquaredX=SquaredX,
                    FixNodesAtPoints=FixNodesAtPoints,
                    PseudotimeNodePositions=PseudotimeNodePositions,
                    PseudotimeLambda=pseudotimeLambda,
                    partition=newpartition,
                    dists=newdists,
                    precomp_d=newprecomp_d,
                )
            else:
                (
                    nodep,
                    ElasticEnergy,
                    part,
                    dist,
                    mse,
                    ep,
                    rp,
                ) = PrimitiveElasticGraphEmbedment_cp_v2(
                    X,
                    FitNodePositions,
                    FitElasticMatrix,
                    MaxNumberOfIterations,
                    eps,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,
                    prob=EmbPointProb,
                    DisplayWarnings=DisplayWarnings,
                    PointWeights=PointWeights,
                    MaxBlockSize=MaxBlockSize,
                    verbose=False,
                    TrimmingRadius=TrimmingRadius,
                    SquaredX=SquaredX,
                    Xcp=Xcp,
                    SquaredXcp=SquaredXcp,
                    FixNodesAtPoints=FixNodesAtPoints,
                    PseudotimeNodePositions=PseudotimeNodePositions,
                    PseudotimeLambda=pseudotimeLambda,
                    partition=newpartition,
                    dists=newdists,
                    precomp_d=newprecomp_d,
                )

            if ElasticEnergy < minEnergy:
                if pseudotime is not None and len(NodePositions) > 10:
                    StoreMergedElasticMatrix = MergedElasticMatrix
                    StoreMergedNodePositions = np.concatenate(
                        (PseudotimeNodePositions, nodep)
                    )
                    NewNodePositions = nodep
                else:
                    StoreMergedElasticMatrix = None
                    StoreMergedNodePositions = None
                    NewNodePositions = nodep
                NewElasticMatrix = ElasticMatricesAll[i]
                partition = part
                AdjustVect = AdjustVectAll[i]
                minEnergy = ElasticEnergy
                MSE = mse
                EP = ep
                RP = rp
                Dist = dist

    # if ~np.isfinite(minEnergy):
    #    return "failed operation"
    return dict(
        StoreMergedElasticMatrix=StoreMergedElasticMatrix,
        StoreMergedNodePositions=StoreMergedNodePositions,
        NodePositions=NewNodePositions,
        ElasticMatrix=NewElasticMatrix,
        ElasticEnergy=minEnergy,
        MSE=MSE,
        EP=EP,
        RP=RP,
        AdjustVect=AdjustVect,
        Dist=Dist,
    )


def ApplyOptimalGraphGrammarOperation_v3(
    X,
    NodePositions,
    ElasticMatrix,
    opTypes,
    AdjustVect=None,
    SquaredX=None,
    verbose=False,
    MaxBlockSize=100000000,
    MaxNumberOfIterations=100,
    eps=0.01,
    TrimmingRadius=float("inf"),
    Mode=1,
    FinalEnergy="Base",
    alpha=1,
    beta=1,
    EmbPointProb=1,
    PointWeights=None,
    AvoidSolitary=False,
    AdjustElasticMatrix=None,
    DisplayWarnings=True,
    n_cores=1,
    MinParOp=20,
    multiproc_shared_variables=None,
    Xcp=None,
    SquaredXcp=None,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    # Multiple grammar application --------------------------------------------
    Application of the grammar operation. This in an internal function that should not be used in by the end-user

    X numerical 2D matrix, the n-by-m matrix with the position of n m-dimensional points
    NodePositions numerical 2D matrix, the k-by-m matrix with the position of k m-dimensional points
    ElasticMatrix numerical 2D matrix, the k-by-k elastic matrix
    operationtypes string vector containing the operation to use
    SquaredX rowSums(X^2), if NULL it will be computed
    verbose boolean. Should addition information be displayed
    n.cores integer. How many cores to use. If EnvCl is not NULL, that cliuster setup will be used,
    otherwise a SOCK cluster willbe used
    EnvCl a cluster structure returned, e.g., by makeCluster.
    If a cluster structure is used, all the nodes must be able to access all the variable needed by PrimitiveElasticGraphEmbedment
    MaxNumberOfIterations is an integer number indicating the maximum number of iterations for the EM algorithm
    TrimmingRadius is a real value indicating the trimming radius, a parameter required for robust principal graphs
    (see https://github.com/auranic/Elastic-principal-graphs/wiki/Robust-principal-graphs)
    eps a real number indicating the minimal relative change in the nodenpositions
    to be considered the graph embedded (convergence criteria)
    Mode integer, the energy mode. It can be 1 (difference is computed using the position of the nodes) and
    2 (difference is computed using the changes in elestic energy of the configuraztions)
    FinalEnergy string indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha positive numeric, the value of the alpha parameter of the penalized elastic energy
    beta positive numeric, the value of the beta parameter of the penalized elastic energy
    gamma
    FastSolve boolean, should FastSolve be used when fitting the points to the data?
    AvoidSolitary boolean, should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    EmbPointProb numeric between 0 and 1. If less than 1 point will be sampled at each iteration. Prob indicate the probability of
    using each points. This is an *experimental* feature, which may helps speeding up the computation if a large number of points is present.
    AdjustVect
    AdjustElasticMatrix
    ...
    MinParOp integer, the minimum number of operations to use parallel computation

    @return

    @examples
    """

    NodePositionsArrayAll = []
    ElasticMatricesAll = []
    AdjustVectAll = []
    NodeIndicesArrayAll = []
    opTypesAll = []

    if Xcp is None:
        partition, _ = PartitionData(
            X, NodePositions, MaxBlockSize, SquaredX, TrimmingRadius
        )
    else:
        partition, _ = PartitionData_cp(
            Xcp, NodePositions, MaxBlockSize, SquaredXcp, TrimmingRadius
        )

    for i in range(len(opTypes)):
        if verbose:
            print(" Operation type : ", opTypes[i])
        (
            NodePositionsArray,
            ElasticMatrices,
            AdjustVectArray,
            NodeIndicesArray,
        ) = GraphGrammarOperation(
            X,
            NodePositions,
            ElasticMatrix,
            AdjustVect,
            opTypes[i],
            partition,
            FixNodesAtPoints,
            MaxNumberOfGraphCandidatesDict,
        )

        NodePositionsArrayAll.extend(NodePositionsArray)
        ElasticMatricesAll.extend(ElasticMatrices)
        AdjustVectAll.extend(AdjustVectArray)
        NodeIndicesArrayAll.extend(NodeIndicesArray.T)
        opTypesAll.extend([opTypes[i]] * len(NodePositionsArray))

    if verbose:
        print("Optimizing graphs")

    Valid_configurations = range(len(NodePositionsArrayAll))

    if AvoidSolitary:
        Valid_configurations = []
        if Xcp is None:
            for i in range(len(NodePositionsArrayAll)):
                partition = PartitionData(
                    X=X,
                    MaxBlockSize=MaxBlockSize,
                    NodePositions=NodePositionsArrayAll[i],
                    SquaredX=SquaredX,
                    TrimmingRadius=TrimmingRadius,
                )[0]
                if all(
                    np.isin(
                        np.array(range(NodePositionsArrayAll[i].shape[0])),
                        partition[partition > -1],
                    )
                ):
                    Valid_configurations.append(i)

        else:
            for i in range(len(NodePositionsArrayAll)):
                partition = PartitionData_cp(
                    Xcp,
                    MaxBlockSize,
                    NodePositionsArrayAll[i],
                    SquaredXcp,
                    TrimmingRadius=TrimmingRadius,
                )[0]
                if all(
                    np.isin(
                        np.array(range(NodePositionsArrayAll[i].shape[0])),
                        partition[partition > -1],
                    )
                ):
                    Valid_configurations.append(i)

        if verbose:
            print(
                len(Valid_configurations),
                "configurations out of ",
                len(NodePositionsArrayAll),
                "used",
            )
        if Valid_configurations == []:
            return "failed operation"
    #     NodePositionArrayAll = NodePositionArrayAll[...,Valid_configurations]
    #     ElasticMatricesAll = ElasticMatricesAll[...,Valid_configurations]
    #     AdjustVectAll = AdjustVectAll[Valid_configurations]

    if AdjustElasticMatrix:
        for i in Valid_configurations:
            ElasticMatricesAll[i], AdjustVectAll[i] = AdjustByConstant(
                ElasticMatricesAll[i], AdjustVectAll[i]
            )

    if n_cores > 1 and len(Valid_configurations) // (MinParOp + 1) > 1:

        #             X_remote, X_shape, SquaredX_remote, SquaredX_shape = multiproc_shared_variables
        #             ####### Multiprocessing
        #             with Pool(processes=n_cores,
        #                       initializer=init_worker,
        #                       initargs=(X_remote,
        #                                 X_shape,
        #                                 SquaredX_remote,
        #                                 SquaredX_shape,
        #                                 NodePositionsArrayAll,
        #                                 ElasticMatricesAll,
        #                                 MaxNumberOfIterations,
        #                                 TrimmingRadius,
        #                                 eps,
        #                                 Mode,
        #                                 FinalEnergy,
        #                                 alpha,
        #                                 beta,
        #                                 EmbPointProb,
        #                                 DisplayWarnings,
        #                                 None,
        #                                 MaxBlockSize,
        #                                 False)) as pool:

        #                 results = pool.map(proxy_multiproc,Valid_configurations)

        with mp.Pool(n_cores) as pool:
            if Xcp is None:
                results = pool.map(
                    proxy,
                    [
                        dict(
                            X=X,
                            NodePositions=NodePositionsArrayAll[i],
                            ElasticMatrix=ElasticMatricesAll[i],
                            MaxNumberOfIterations=MaxNumberOfIterations,
                            eps=eps,
                            Mode=Mode,
                            FinalEnergy=FinalEnergy,
                            alpha=alpha,
                            beta=beta,
                            prob=EmbPointProb,
                            DisplayWarnings=DisplayWarnings,
                            PointWeights=PointWeights,
                            MaxBlockSize=MaxBlockSize,
                            verbose=False,
                            TrimmingRadius=TrimmingRadius,
                            SquaredX=SquaredX,
                        )
                        for i in Valid_configurations
                    ],
                )
            else:
                results = pool.map(
                    proxy_cp,
                    [
                        dict(
                            X=X,
                            NodePositions=NodePositionsArrayAll[i],
                            ElasticMatrix=ElasticMatricesAll[i],
                            MaxNumberOfIterations=MaxNumberOfIterations,
                            eps=eps,
                            Mode=Mode,
                            FinalEnergy=FinalEnergy,
                            alpha=alpha,
                            beta=beta,
                            prob=EmbPointProb,
                            DisplayWarnings=DisplayWarnings,
                            PointWeights=PointWeights,
                            MaxBlockSize=MaxBlockSize,
                            verbose=False,
                            TrimmingRadius=TrimmingRadius,
                            SquaredX=SquaredX,
                            Xcp=Xcp,
                            SquaredXcp=SquaredXcp,
                        )
                        for i in Valid_configurations
                    ],
                )

        list_energies = [r[1] for r in results]
        idx = list_energies.index(min(list_energies))
        NewNodePositions, minEnergy, partition, Dist, MSE, EP, RP = results[
            idx
        ]
        AdjustVect = AdjustVectAll[idx]
        NewElasticMatrix = ElasticMatricesAll[idx]

    else:
        minEnergy = np.inf
        StoreMeanPseudotime = None
        StoreMergedElasticMatrix = None
        StoreMergedNodePositions = None
        if Xcp is None:
            for i in Valid_configurations:
                # TODO add pointweights ?
                if pseudotime is not None and len(NodePositions) > 10:
                    (
                        MeanPseudotime,
                        MergedNodePositions,
                        MergedElasticMatrix,
                        MergedEdges,
                        nPseudoNodes,
                    ) = gen_pseudotime_augmented_graph_by_path(
                        X,
                        SquaredX,
                        NodePositionsArrayAll[i],
                        ElasticMatricesAll[i],
                        pseudotime,
                        root_node=0,
                        LinkMu=0,
                        LinkLambda=pseudotimeLambda,
                        PointWeights=PointWeights,
                        TrimmingRadius=TrimmingRadius,
                    )

                    _split = int(len(MergedNodePositions) / 2)
                    FitNodePositions = MergedNodePositions[_split:]
                    PseudotimeNodePositions = MergedNodePositions[:_split]
                    FitElasticMatrix = ElasticMatricesAll[i]

                else:
                    FitNodePositions = NodePositionsArrayAll[i]
                    FitElasticMatrix = ElasticMatricesAll[i]
                    PseudotimeNodePositions = None
                    nPseudoNodes = 0
                (
                    nodep,
                    ElasticEnergy,
                    part,
                    dist,
                    mse,
                    ep,
                    rp,
                ) = PrimitiveElasticGraphEmbedment_v2(
                    X,
                    FitNodePositions,
                    FitElasticMatrix,
                    MaxNumberOfIterations,
                    eps,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,
                    prob=EmbPointProb,
                    DisplayWarnings=DisplayWarnings,
                    PointWeights=PointWeights,
                    MaxBlockSize=MaxBlockSize,
                    verbose=False,
                    TrimmingRadius=TrimmingRadius,
                    SquaredX=SquaredX,
                    FixNodesAtPoints=FixNodesAtPoints,
                    PseudotimeNodePositions=PseudotimeNodePositions,
                    PseudotimeLambda=pseudotimeLambda,
                    # Label=label,
                    # LabelLambda=labelLambda,
                )

                if ElasticEnergy < minEnergy:
                    if pseudotime is not None and len(NodePositions) > 10:
                        StoreMeanPseudotime = MeanPseudotime
                        StoreMergedElasticMatrix = MergedElasticMatrix
                        StoreMergedNodePositions = np.concatenate(
                            (PseudotimeNodePositions, nodep)
                        )
                        NewNodePositions = nodep
                    else:
                        StoreMeanPseudotime = None
                        StoreMergedElasticMatrix = None
                        StoreMergedNodePositions = None
                        NewNodePositions = nodep
                    NewElasticMatrix = ElasticMatricesAll[i]
                    partition = part
                    AdjustVect = AdjustVectAll[i]
                    minEnergy = ElasticEnergy
                    MSE = mse
                    EP = ep
                    RP = rp
                    Dist = dist

        else:
            for i in Valid_configurations:
                # TODO add pointweights ?
                if pseudotime is not None and len(NodePositions) > 10:
                    (
                        MeanPseudotime,
                        MergedNodePositions,
                        MergedElasticMatrix,
                        MergedEdges,
                        nPseudoNodes,
                    ) = gen_pseudotime_augmented_graph_by_path(
                        X,
                        SquaredX,
                        NodePositionsArrayAll[i],
                        ElasticMatricesAll[i],
                        pseudotime,
                        root_node=0,
                        LinkMu=0,
                        LinkLambda=pseudotimeLambda,
                        PointWeights=PointWeights,
                        TrimmingRadius=TrimmingRadius,
                    )
                    _split = int(len(MergedNodePositions) / 2)
                    FitNodePositions = MergedNodePositions[_split:]
                    PseudotimeNodePositions = MergedNodePositions[:_split]
                    FitElasticMatrix = ElasticMatricesAll[i]
                else:
                    FitNodePositions = NodePositionsArrayAll[i]
                    FitElasticMatrix = ElasticMatricesAll[i]
                    PseudotimeNodePositions = None
                    nPseudoNodes = 0
                (
                    nodep,
                    ElasticEnergy,
                    part,
                    dist,
                    mse,
                    ep,
                    rp,
                ) = PrimitiveElasticGraphEmbedment_cp_v2(
                    X,
                    FitNodePositions,
                    FitElasticMatrix,
                    MaxNumberOfIterations,
                    eps,
                    Mode=Mode,
                    FinalEnergy=FinalEnergy,
                    alpha=alpha,
                    beta=beta,
                    prob=EmbPointProb,
                    DisplayWarnings=DisplayWarnings,
                    PointWeights=PointWeights,
                    MaxBlockSize=MaxBlockSize,
                    verbose=False,
                    TrimmingRadius=TrimmingRadius,
                    SquaredX=SquaredX,
                    Xcp=Xcp,
                    SquaredXcp=SquaredXcp,
                    FixNodesAtPoints=FixNodesAtPoints,
                    PseudotimeNodePositions=PseudotimeNodePositions,
                    PseudotimeLambda=pseudotimeLambda,
                    # Label=label,
                    # LabelLambda=labelLambda,
                )

                if ElasticEnergy < minEnergy:
                    if pseudotime is not None and len(NodePositions) > 10:
                        StoreMeanPseudotime = MeanPseudotime
                        StoreMergedElasticMatrix = MergedElasticMatrix
                        StoreMergedNodePositions = np.concatenate(
                            (PseudotimeNodePositions, nodep)
                        )
                        NewNodePositions = nodep
                    else:
                        StoreMeanPseudotime = None
                        StoreMergedElasticMatrix = None
                        StoreMergedNodePositions = None
                        NewNodePositions = nodep
                    NewElasticMatrix = ElasticMatricesAll[i]
                    partition = part
                    AdjustVect = AdjustVectAll[i]
                    minEnergy = ElasticEnergy
                    MSE = mse
                    EP = ep
                    RP = rp
                    Dist = dist

    # if ~np.isfinite(minEnergy):
    #    return "failed operation"
    return dict(
        StoreMeanPseudotime=StoreMeanPseudotime,
        StoreMergedElasticMatrix=StoreMergedElasticMatrix,
        StoreMergedNodePositions=StoreMergedNodePositions,
        NodePositions=NewNodePositions,
        ElasticMatrix=NewElasticMatrix,
        ElasticEnergy=minEnergy,
        MSE=MSE,
        EP=EP,
        RP=RP,
        AdjustVect=AdjustVect,
        Dist=Dist,
    )



from mimetypes import MimeTypes
import networkx as nx
import elpigraph
import numpy as np
import numba as nb
import matplotlib.pyplot as plt
import itertools
import pandas as pd
from sklearn.decomposition import PCA
from copy import deepcopy
from scipy.spatial.qhull import ConvexHull
from shapely.geometry import Point, Polygon, MultiLineString, LineString
from shapely.geometry.multipolygon import MultiPolygon
from sklearn.neighbors import NearestNeighbors


def flatten(S):
    if S == []:
        return S
    if isinstance(S[0], list):
        return flatten(S[0]) + flatten(S[1:])
    return S[:1] + flatten(S[1:])


@nb.njit
def _get_intersect_inner(a1, a2, b1, b2):
    """
    Returns the point of intersection of the lines passing through a2,a1 and b2,b1.
    a1: [x, y] a point on the first line
    a2: [x, y] another point on the first line
    b1: [x, y] a point on the second line
    b2: [x, y] another point on the second line
    """
    s = np.vstack((a1, a2, b1, b2))  # s for stacked
    h = np.hstack((s, np.ones((4, 1))))  # h for homogeneous
    l1 = np.cross(h[0], h[1])  # get first line
    l2 = np.cross(h[2], h[3])  # get second line
    x, y, z = np.cross(l1, l2)  # point of intersection
    if z == 0:  # lines are parallel
        return np.array([np.inf, np.inf])
    return np.array([x / z, y / z])


@nb.njit
def _isBetween(a, b, c):
    """ Check if c is in between a and b """
    dotproduct = (c[0] - a[0]) * (b[0] - a[0]) + (c[1] - a[1]) * (b[1] - a[1])
    if dotproduct < 0:
        return False

    squaredlengthba = (b[0] - a[0]) * (b[0] - a[0]) + (b[1] - a[1]) * (b[1] - a[1])
    if dotproduct > squaredlengthba:
        return False
    return True


@nb.njit
def _get_intersect(Xin, nodep, edges, cent):
    inters = np.zeros_like(Xin)
    for i, x in enumerate(Xin):
        for e in edges:
            p_inter = _get_intersect_inner(nodep[e[0]], nodep[e[1]], x, cent)
            if _isBetween(nodep[e[0]], nodep[e[1]], p_inter):
                if np.sum((x - p_inter) ** 2) < np.sum((cent - p_inter) ** 2):
                    inters[i] = p_inter
    return inters


@nb.njit
def get_weights_lineproj(Xin, nodep, edges, cent, threshold=0.2):

    Xin_lineproj = _get_intersect(Xin, nodep, edges, cent)
    distcent_Xin_lineproj = np.sqrt(np.sum((Xin_lineproj - cent) ** 2, axis=1))
    distcent_Xin = np.sqrt(np.sum((Xin - cent) ** 2, axis=1))

    w = 1 - distcent_Xin / distcent_Xin_lineproj
    idx_close = w > threshold
    w[idx_close] = 1.0
    return w, idx_close


def shrink_or_swell_shapely_polygon(coords, factor=0.10, swell=False):
    """returns the shapely polygon which is smaller or bigger by passed factor.
    If swell = True , then it returns bigger polygon, else smaller"""

    my_polygon = Polygon(coords)
    xs = list(my_polygon.exterior.coords.xy[0])
    ys = list(my_polygon.exterior.coords.xy[1])
    x_center = 0.5 * min(xs) + 0.5 * max(xs)
    y_center = 0.5 * min(ys) + 0.5 * max(ys)
    min_corner = Point(min(xs), min(ys))
    max_corner = Point(max(xs), max(ys))
    center = Point(x_center, y_center)
    shrink_distance = center.distance(min_corner) * factor

    if swell:
        my_polygon_resized = my_polygon.buffer(shrink_distance)  # expand
    else:
        my_polygon_resized = my_polygon.buffer(-shrink_distance)  # shrink

    return my_polygon_resized


def remove_intersections(nodep, edges):
    """ Update edges to account for possible 2d intersections in graph after adding loops """
    new_nodep = nodep.copy()
    lnodep = new_nodep.tolist()
    new_edges = edges.tolist()
    multiline = MultiLineString([LineString(new_nodep[e]) for e in new_edges])

    while not (multiline.is_simple):  # while intersections in graph

        # find an intersection, update edges, break, update graph
        for i, j in itertools.combinations(range(len(multiline)), 2):
            line1, line2 = multiline[i], multiline[j]
            if line1.intersects(line2):
                if list(np.array(line1.intersection(line2))) not in lnodep:
                    new_nodep = np.append(
                        new_nodep, np.array(line1.intersection(line2))[None], axis=0,
                    )
                    intersects_idx = [list(new_edges[i]), list(new_edges[j])]
                    new_edges.pop(new_edges.index(intersects_idx[0]))
                    new_edges.pop(new_edges.index(intersects_idx[1]))

                    for n in np.array(intersects_idx).flatten():
                        new_edges.append([n, len(new_nodep) - 1])
                    break

        multiline = MultiLineString([LineString(new_nodep[e]) for e in new_edges])
        lnodep = new_nodep.tolist()
    return new_nodep, np.array(new_edges)


@nb.njit
def mahalanobis(M, cent):
    cov = np.cov(M, rowvar=0)
    try:
        cov_inverse = np.linalg.inv(cov)
    except:
        cov_inverse = np.linalg.pinv(cov)

    M_c = M - cent
    dist = np.sqrt(np.sum((M_c) * cov_inverse.dot(M_c.T).T, axis=1))
    return dist


@nb.njit
def polygon_area(x, y):
    # coordinate shift
    x_ = x - x.mean()
    y_ = y - y.mean()
    # everything else is the same as maxb's code
    correction = x_[-1] * y_[0] - y_[-1] * x_[0]
    main_area = np.dot(x_[:-1], y_[1:]) - np.dot(y_[:-1], x_[1:])
    return 0.5 * np.abs(main_area + correction)


def pp_compactness(cycle_nodep):
    """Polsby-Popper compactness"""
    area = polygon_area(cycle_nodep[:, 0], cycle_nodep[:, 1])
    length = np.sum(np.sqrt((np.diff(cycle_nodep, axis=0) ** 2).sum(axis=1)))
    return (4 * np.pi * area) / (length ** 2)


def find_all_cycles(G, source=None, cycle_length_limit=None):
    """forked from networkx dfs_edges function. Assumes nodes are integers, or at least
    types which work with min() and > ."""
    if source is None:
        # produce edges for all components
        nodes = [list(i)[0] for i in nx.connected_components(G)]
    else:
        # produce edges for components with source
        nodes = [source]
    # extra variables for cycle detection:
    cycle_stack = []
    output_cycles = set()

    def get_hashable_cycle(cycle):
        """cycle as a tuple in a deterministic order."""
        m = min(cycle)
        mi = cycle.index(m)
        mi_plus_1 = mi + 1 if mi < len(cycle) - 1 else 0
        if cycle[mi - 1] > cycle[mi_plus_1]:
            result = cycle[mi:] + cycle[:mi]
        else:
            result = list(reversed(cycle[:mi_plus_1])) + list(
                reversed(cycle[mi_plus_1:])
            )
        return tuple(result)

    for start in nodes:
        if start in cycle_stack:
            continue
        cycle_stack.append(start)

        stack = [(start, iter(G[start]))]
        while stack:
            parent, children = stack[-1]
            try:
                child = next(children)

                if child not in cycle_stack:
                    cycle_stack.append(child)
                    stack.append((child, iter(G[child])))
                else:
                    i = cycle_stack.index(child)
                    if i < len(cycle_stack) - 2:
                        output_cycles.add(get_hashable_cycle(cycle_stack[i:]))

            except StopIteration:
                stack.pop()
                cycle_stack.pop()

    return [list(i) for i in output_cycles]


# def in_hull(points, queries):
#    hull = _Qhull(
#        b"i",
#        points,
#        options=b"",
#        furthest_site=False,
#        incremental=False,
#        interior_point=None,
#    )
#    equations = hull.get_simplex_facet_array()[2].T
#    return np.all(queries @ equations[:-1] < -equations[-1], axis=1)


def in_hull(points, queries):
    equations = ConvexHull(points).equations.T
    return np.all(queries @ equations[:-1] < -equations[-1], axis=1)


def findPaths(
    X,
    PG,
    min_path_len=None,
    nnodes=None,
    max_inner_fraction=0.1,
    min_node_n_points=2,
    max_n_points=None,
    # max_empty_curve_fraction=.2,
    min_compactness=0.5,
    radius=None,
    allow_same_branch=True,
    fit_loops=True,
    Lambda=None,
    Mu=None,
    cycle_Lambda=None,
    cycle_Mu=None,
    weights=None,
    plot=False,
    verbose=False,
):
    """
    This function tries to add extra paths to the graph
    by computing a series of principal curves connecting two nodes
    and retaining plausible ones using heuristic parameters

    min_path_len: int, default=None
        Minimum distance along the graph (in number of nodes) that separates the two nodes to connect with a principal curve
    n_nodes: int, default=None
        Number of nodes in the candidate principal curves
    max_inner_fraction: float in [0,1], default=0.1
        Maximum fraction of points inside vs outside the loop Controls how empty the loop formed with the added path should be.
    min_node_n_points: int, default=1
        Minimum number of points associated to nodes of the principal curve (prevents creating paths through empty space)
    max_n_points: int, default=5% of the number of points
        Maximum number of points inside the loop
    min_compactness: float in [0,1], default=0.5
        Minimum 'roundness' of the loop (1=more round) (if very narrow loops are not desired)
    radius: float, default=None
        Max distance in space that separates the two nodes to connect with a principal curve
    allow_same_branch: bool, default=True
        Whether to allow new paths to connect two nodes from the same branch
    fit_loops: bool, default=True
        Whether to refit the graph to data after adding the new paths
    plot: bool, default=False
        Whether to plot selected candidate paths
    verbose: bool, default=False
    copy: bool, default=False
    use_weights: bool, default=False
        Whether to use point weights
    use_partition: bool or list, default=False
    """

    _PG = deepcopy(PG)
    if "projection" in _PG.keys():
        del _PG["projection"]
    init_nodes_pos, init_edges = _PG["NodePositions"], _PG["Edges"][0]

    # --- Init parameters, variables
    epg = nx.Graph(init_edges.tolist())

    SquaredX = np.sum(X ** 2, axis=1, keepdims=1)
    part, part_dist = elpigraph.src.core.PartitionData(
        X, init_nodes_pos, 10 ** 6, SquaredX=SquaredX
    )
    leaves = [k for k, v in epg.degree if v == 1]
    edge_lengths = np.sqrt(
        np.sum(
            (init_nodes_pos[init_edges[:, 0], :] - init_nodes_pos[init_edges[:, 1], :])
            ** 2,
            axis=1,
        )
    )

    if Mu is None:
        Mu = _PG["Mu"]
    if Lambda is None:
        Lambda = _PG["Lambda"]
    if cycle_Mu is None:
        cycle_Mu = Mu / 10
    if cycle_Lambda is None:
        cycle_Lambda = Lambda / 10
    if radius is None:
        radius = np.mean(edge_lengths) * len(init_nodes_pos) / 10
    if min_path_len is None:
        min_path_len = len(init_nodes_pos) // 5
    if max_n_points is None:
        max_n_points = int(len(X) * 0.05)
    if min_node_n_points is None:
        min_node_n_points = max(1, np.bincount(part.flat).min())
    if weights is None:
        weights = np.ones(len(X))[:, None]
    if nnodes is None:
        nnodes = min(16, max(6, int(radius / np.mean(edge_lengths))))
    elif nnodes < 6:
        raise ValueError("nnodes should be at least 6")

    if verbose:
        print(
            f"Using default parameters: max_n_points={max_n_points},"
            f" radius={radius:.2f}, min_node_n_points={min_node_n_points},"
            f" min_path_len={min_path_len}, nnodes={nnodes}"
        )

    # --- Get candidate nodes to connect
    dist, ind = (
        NearestNeighbors(radius=radius)
        .fit(init_nodes_pos)
        .radius_neighbors(init_nodes_pos[leaves])
    )
    net = elpigraph.src.graphs.ConstructGraph({"Edges": [init_edges]})

    if all(np.array(net.degree()) <= 2):
        branches = net.get_shortest_paths(leaves[0], leaves[-1])
    else:
        (
            dict_tree,
            dict_branches,
            dict_branches_single_end,
        ) = elpigraph.src.supervised.get_tree(init_edges, leaves[0])
        branches = list(dict_branches.values())

    candidate_nodes = []
    for i in range(len(leaves)):
        root_branch = [b for b in branches if leaves[i] in b][0]

        if allow_same_branch:
            _cand_nodes = [node for b in branches for node in b if node in ind[i]]
        else:
            _cand_nodes = [
                node
                for b in branches
                for node in b
                if not (node in root_branch) and node in ind[i]
            ]
        paths = net.get_shortest_paths(leaves[i], _cand_nodes)
        candidate_nodes.append([p[-1] for p in paths if len(p) > min_path_len])

    # --- Test each of the loops connecting a leaf to its candidate nodes,
    # --- for each leaf select the one with minimum energy and that respect constraints
    if verbose:
        print("testing", sum([len(_) for _ in candidate_nodes]), "candidates")

    new_edges = []
    new_nodep = []
    new_leaves = []
    new_part = []
    new_energy = []
    new_inner_fraction = []
    for i, l in enumerate(leaves):
        inner_fractions = []
        energies = []
        merged_edges = []
        merged_nodep = []
        merged_part = []
        loop_edges = []
        loop_nodep = []
        loop_leaves = []
        for c in candidate_nodes[i]:

            clus = (part == c) | (part == l)
            X_fit = np.vstack((init_nodes_pos[c], init_nodes_pos[l], X[clus.flat]))
            try:
                pg = elpigraph.computeElasticPrincipalCurve(
                    X_fit, nnodes, Lambda=Lambda, Mu=Mu, FixNodesAtPoints=[[0], [1]],
                )[0]
            except Exception as e:
                energies.append(np.inf)
                merged_edges.append(np.inf)
                merged_nodep.append(np.inf)
                loop_edges.append(np.inf)
                loop_nodep.append(np.inf)
                loop_leaves.append(np.inf)
                # candidate curve has infinite energy, ignore error
                if e.args == (
                    "local variable 'NewNodePositions' referenced before" " assignment",
                ):
                    continue
                else:
                    raise e

            # ---get nodep, edges, create new graph with added loop
            nodep, edges = pg["NodePositions"], pg["Edges"][0]
            # _part, _part_dist = elpigraph.src.core.PartitionData(
            #    X_fit, nodep, 10 ** 6, np.sum(X_fit ** 2, axis=1, keepdims=1)
            # )
            _edges = edges.copy()
            _edges[(edges != 0) & (edges != 1)] += init_edges.max() - 1
            _edges[edges == 0] = c
            _edges[edges == 1] = l
            _merged_edges = np.concatenate((init_edges, _edges))
            _merged_nodep = np.concatenate((init_nodes_pos, nodep[2:]))

            cycle_nodes = find_all_cycles(nx.Graph(_merged_edges.tolist()))[0]

            ElasticMatrix = elpigraph.src.core.MakeUniformElasticMatrix_with_cycle(
                _merged_edges,
                Lambda=Lambda,
                Mu=Mu,
                cycle_Lambda=cycle_Lambda,
                cycle_Mu=cycle_Mu,
                cycle_nodes=cycle_nodes,
            )

            (
                _merged_nodep,
                _,
                _,
                _,
                _,
                _,
                _,
            ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
                X,
                _merged_nodep,
                ElasticMatrix,
                PointWeights=weights,
                FixNodesAtPoints=[],
            )

            ### candidate validity tests ###
            valid = True
            # --- curve validity test
            # if (max_empty_curve_fraction is not None) and valid: # if X_fit projected to curve has long gaps
            #    infer_pseudotime(_adata,source=0)
            #    sorted_X_proj=_adata.obsm['X_epg_proj'][_adata.obs['epg_pseudotime'].argsort()]
            #    dist = np.sqrt((np.diff(sorted_X_proj,axis=0)**2).sum(axis=1))
            #    curve_len = np.sum(_adata.uns['epg']['edge_len'])
            #    if np.max(dist) > (curve_len * max_empty_curve_fraction):
            #        valid = False

            # --- cycle validity test
            if valid:
                G = nx.Graph(_merged_edges.tolist())
                cycle_nodes = find_all_cycles(G)[0]
                cycle_nodep = np.array([_merged_nodep[e] for e in cycle_nodes])
                cent_part, cent_dists = elpigraph.src.core.PartitionData(
                    X, _merged_nodep, 10 ** 6, SquaredX=SquaredX
                )
                cycle_points = np.isin(cent_part.flat, cycle_nodes)

                if X.shape[1] > 2:
                    pca = PCA(n_components=2, svd_solver="arpack").fit(X[cycle_points])
                    X_cycle_2d = pca.transform(X[cycle_points])
                    cycle_2d = pca.transform(cycle_nodep)
                else:
                    cycle_2d = cycle_nodep
                    X_cycle_2d = X[cycle_points]

                inside_idx = in_hull(cycle_2d, X_cycle_2d)

                if sum(inside_idx) == 0:
                    inner_fraction = 0.0
                else:
                    cycle_centroid = np.mean(cycle_2d, axis=0, keepdims=1)
                    X_inside = X_cycle_2d[inside_idx]

                    if len(X_inside) == 1:
                        w = np.ones(len(X_inside))
                    else:
                        w = mahalanobis(X_inside, cycle_centroid)

                    # points belonging to cycle shrunk by 10% or within 2 std of centroid (mahalanobis < 2)
                    shrunk_cycle_2d = shrink_or_swell_shapely_polygon(
                        cycle_2d, factor=0.1
                    )

                    # prevent shapely bugs when multi-polygon is returned. Fall back to mahalanobis
                    if type(shrunk_cycle_2d) == MultiPolygon:
                        in_shrunk_cycle = np.ones(len(X_inside), dtype=bool)
                    else:
                        shrunk_cycle_2d = np.array(shrunk_cycle_2d.exterior.coords)

                        # prevent bug when self-intersection
                        if len(shrunk_cycle_2d) == 0:
                            in_shrunk_cycle = np.ones(len(X_inside), dtype=bool)
                        else:
                            in_shrunk_cycle = in_hull(shrunk_cycle_2d, X_inside)
                    idx_close = in_shrunk_cycle | (w < 1)
                    w = 1 - w / w.max()
                    w[idx_close] = 1

                    # cycle_nodes_array = np.append(np.array(list(zip(range(len(cycle_2d)-1),
                    #                                  range(1,len(cycle_2d))))),[[len(cycle_2d)-1,0]],axis=0)
                    # w, idx_close = get_weights_lineproj(X_inside,cycle_2d,cycle_nodes_array,cycle_centroid[0],threshold=.2)

                    inner_fraction = np.sum(w) / np.sum(cycle_points)

                if init_nodes_pos.shape[1] == 2:
                    intersect = not (
                        MultiLineString(
                            [LineString(_merged_nodep[e]) for e in _merged_edges]
                        ).is_simple
                    )
                    if intersect:
                        valid = False

                # idx for min_node_n_points test: points that are away from the center
                ix_outside = np.ones(len(cycle_points), dtype=bool)
                ix_outside[
                    np.arange(len(X))[cycle_points][inside_idx][idx_close]
                ] = False
                if (
                    any(
                        np.bincount(
                            cent_part[ix_outside].flat, minlength=len(_merged_nodep),
                        )[len(init_nodes_pos) :]
                        < min_node_n_points
                    )  # if empty cycle node
                    or (
                        inner_fraction > max_inner_fraction
                    )  # if high fraction of points inside
                    or (not np.isfinite(inner_fraction))  # prevent no points error
                    or (np.sum(idx_close) > max_n_points)  # if too many points inside
                    or pp_compactness(cycle_2d) < min_compactness
                ):  # if loop is very narrow
                    valid = False

            # ---> if cycle is invalid, continue
            if not valid:
                inner_fractions.append(np.inf)
                energies.append(np.inf)
                merged_edges.append(np.inf)
                merged_nodep.append(np.inf)
                merged_part.append(np.inf)
                loop_edges.append(np.inf)
                loop_nodep.append(np.inf)
                loop_leaves.append(np.inf)
                continue

            # ---> valid cycle, compute graph energy
            else:
                (_merged_part, _merged_part_dist,) = elpigraph.src.core.PartitionData(
                    X, _merged_nodep, 10 ** 6, SquaredX=SquaredX
                )
                proj = elpigraph.src.reporting.project_point_onto_graph(
                    X, _merged_nodep, _merged_edges, _merged_part
                )
                MSE = proj["MSEP"]
                # dist2proj = np.sum(np.square(X - X_proj), axis=1)
                # ElasticMatrix = elpigraph.src.core.Encode2ElasticMatrix(_merged_edges, Lambdas=Lambda, Mus=Mu)
                # ElasticEnergy, MSE, EP, RP = elpigraph.src.core.ComputePenalizedPrimitiveGraphElasticEnergy(_merged_nodep,
                #                                                                                            ElasticMatrix,
                #                                                                                            dist2proj,alpha=0.01,beta=0)
                inner_fractions.append(inner_fraction)
                energies.append(MSE)
                merged_edges.append(_merged_edges)
                merged_nodep.append(_merged_nodep)
                merged_part.append(np.where(np.isin(_merged_part.flat, cycle_nodes))[0])
                loop_edges.append(edges)
                loop_nodep.append(nodep[2:])
                loop_leaves.append([c, l])

        # --- among all valid cycles found, retain the best
        if energies != [] and np.isfinite(energies).any():
            best = np.argmin(energies)
            if [loop_leaves[best][-1], loop_leaves[best][0]] not in new_leaves:
                # and not any(np.isin(loop_leaves[best],np.unique(np.array(new_leaves))))):

                new_edges.append(loop_edges[best])
                new_nodep.append(loop_nodep[best])
                new_leaves.append(loop_leaves[best])
                new_part.append(merged_part[best])
                new_energy.append(energies[best])
                new_inner_fraction.append(inner_fractions[best])
                _merged_edges = merged_edges[best]
                _merged_nodep = merged_nodep[best]

                if plot:
                    c = candidate_nodes[i][best]
                    clus = (part == c) | (part == l)
                    X_fit = np.vstack(
                        (init_nodes_pos[c], init_nodes_pos[l], X[clus.flat])
                    )
                    proj = elpigraph.src.reporting.project_point_onto_graph(
                        X, _merged_nodep, _merged_edges, _merged_part
                    )
                    MSE = proj["MSEP"]

                    # ----- cycle test
                    G = nx.Graph(_merged_edges.tolist())
                    cycle_nodes = find_all_cycles(G)[0]
                    cycle_nodep = np.array([_merged_nodep[e] for e in cycle_nodes])
                    cent_part, cent_dists = elpigraph.src.core.PartitionData(
                        X, _merged_nodep, 10 ** 6, SquaredX=SquaredX
                    )
                    cycle_points = np.isin(cent_part.flat, cycle_nodes)

                    if X.shape[1] > 2:
                        pca = PCA(n_components=2, svd_solver="arpack").fit(cycle_nodep)
                        cycle_2d = pca.transform(cycle_nodep)
                        X_cycle_2d = pca.transform(X[cycle_points])
                    else:
                        cycle_2d = cycle_nodep
                        X_cycle_2d = X[cycle_points]
                    inside_idx = in_hull(cycle_2d, X_cycle_2d)

                    cycle_centroid = np.mean(cycle_2d, axis=0, keepdims=1)
                    X_inside = X_cycle_2d[inside_idx]

                    if len(X_inside) == 1:
                        w = np.ones(len(X_inside))
                    else:
                        w = mahalanobis(X_inside, cycle_centroid)

                    # points belonging to cycle shrunk by 10% or within 2 std of centroid (mahalanobis < 2)
                    shrunk_cycle_2d = shrink_or_swell_shapely_polygon(
                        cycle_2d, factor=0.1
                    )

                    # prevent shapely bugs when multi-polygon is returned. Fall back to mahalanobis
                    if type(shrunk_cycle_2d) == MultiPolygon:
                        in_shrunk_cycle = np.ones(len(X_inside), dtype=bool)
                    else:
                        shrunk_cycle_2d = np.array(shrunk_cycle_2d.exterior.coords)

                        # prevent bug when self-intersection
                        if len(shrunk_cycle_2d) == 0:
                            in_shrunk_cycle = np.ones(len(X_inside), dtype=bool)
                        else:
                            in_shrunk_cycle = in_hull(shrunk_cycle_2d, X_inside)
                    idx_close = in_shrunk_cycle | (w < 1)
                    w = 1 - w / w.max()
                    w[idx_close] = 1
                    inner_fraction = np.sum(w) / np.sum(cycle_points)

                    compactness = pp_compactness(cycle_2d)

                    plt.title(
                        f"{c}, {l}, MSE={MSE:.4f}, \n"
                        f" inner%={inner_fraction:.2f},"
                        f" compactness={compactness:.2f}"
                    )
                    plt.scatter(*X[:, :2].T, alpha=0.1, s=5)
                    plt.scatter(*X_fit[:, :2].T, s=5)
                    plt.scatter(*_merged_nodep[:, :2].T, c="k")
                    for e in _merged_edges:
                        plt.plot(
                            [_merged_nodep[e[0], 0], _merged_nodep[e[1], 0]],
                            [_merged_nodep[e[0], 1], _merged_nodep[e[1], 1]],
                            c="k",
                        )

                    _ = plt.scatter(*X[cycle_points][inside_idx, :2].T, c=w.flat, s=5)
                    plt.colorbar(_)

                    plt.show()

    # ignore equivalent loops (with more than 2/3 shared points and nodes)
    valid = np.ones(len(new_part))
    for i in range(len(new_part) - 1):
        for j in range(i + 1, len(new_part)):
            if (
                len(np.intersect1d(new_part[i], new_part[j]))
                / max(len(new_part[i]), len(new_part[j]))
            ) > (2 / 3):
                if np.argmin([new_inner_fraction[i], new_inner_fraction[j]]) == 0:
                    valid[i] = 0
                else:
                    valid[j] = 0

    new_edges = [e for i, e in enumerate(new_edges) if valid[i]]
    new_nodep = [e for i, e in enumerate(new_nodep) if valid[i]]
    new_leaves = [e for i, e in enumerate(new_leaves) if valid[i]]
    new_part = [e for i, e in enumerate(new_part) if valid[i]]
    new_energy = [e for i, e in enumerate(new_energy) if valid[i]]
    new_inner_fraction = [e for i, e in enumerate(new_inner_fraction) if valid[i]]

    ### form graph with all valid loops found ###
    if (new_edges == []) or (sum(valid) == 0):
        print("Found no valid path to add")
        return None

    for i, loop_edges in enumerate(new_edges):
        if i == 0:
            loop_edges[(loop_edges != 0) & (loop_edges != 1)] += init_edges.max() - 1
            loop_edges[loop_edges == 0] = new_leaves[i][0]
            loop_edges[loop_edges == 1] = new_leaves[i][1]
            merged_edges = np.concatenate((init_edges, loop_edges))
        else:
            loop_edges[(loop_edges != 0) & (loop_edges != 1)] += merged_edges.max() - 1
            loop_edges[loop_edges == 0] = new_leaves[i][0]
            loop_edges[loop_edges == 1] = new_leaves[i][1]
            merged_edges = np.concatenate((merged_edges, loop_edges))
    merged_nodep = np.concatenate((init_nodes_pos, *new_nodep))

    ### optionally refit the entire graph ###
    if fit_loops:
        cycle_nodes = np.concatenate(find_all_cycles(nx.Graph(merged_edges.tolist())))

        ElasticMatrix = elpigraph.src.core.MakeUniformElasticMatrix_with_cycle(
            merged_edges,
            Lambda=Lambda,
            Mu=Mu,
            cycle_Lambda=cycle_Lambda,
            cycle_Mu=cycle_Mu,
            cycle_nodes=cycle_nodes,
        )

        (
            merged_nodep,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
            X, merged_nodep, ElasticMatrix, PointWeights=weights, FixNodesAtPoints=[],
        )
        # check intersection
        if merged_nodep.shape[1] == 2:
            intersect = not (
                MultiLineString(
                    [LineString(merged_nodep[e]) for e in merged_edges]
                ).is_simple
            )
            if intersect:
                merged_nodep, merged_edges = remove_intersections(
                    merged_nodep, merged_edges
                )

    _PG["Edges"] = [merged_edges]
    _PG["NodePositions"] = merged_nodep
    _PG["Lambda"] = Lambda
    _PG["Mu"] = Mu
    _PG["cycle_Lambda"] = cycle_Lambda
    _PG["cycle_Mu"] = cycle_Mu
    _PG["addLoopsdict"] = dict(
        new_edges=new_edges,
        new_nodep=new_nodep,
        new_leaves=new_leaves,
        new_part=new_part,
        new_energy=new_energy,
        new_inner_fraction=new_inner_fraction,
    )

    if verbose:
        l = [
            _PG["addLoopsdict"]["new_inner_fraction"],
            _PG["addLoopsdict"]["new_energy"],
            [len(n) for n in _PG["addLoopsdict"]["new_part"]],
        ]
        df = pd.concat(
            [pd.DataFrame(_PG["addLoopsdict"]["new_leaves"])]
            + [pd.Series(i) for i in l],
            axis=1,
        )
        df.columns = [
            "source node",
            "target node",
            "inner fraction",
            "MSE",
            "n° of points in path",
        ]
        df.index = ["" for i in df.index]
        print("Suggested paths:")
        print(df.round(4))
    return _PG


def addPath(
    X,
    PG,
    source,
    target,
    n_nodes=None,
    weights=None,
    refit_graph=False,
    Mu=None,
    Lambda=None,
    cycle_Mu=None,
    cycle_Lambda=None,
):
    """
    Add path (principal curve) between two nodes in the principal graph

    """

    _PG = deepcopy(PG)
    if "projection" in _PG.keys():
        del _PG["projection"]

    init_nodes_pos, init_edges = _PG["NodePositions"], _PG["Edges"][0]
    init_edges_max = init_edges.max()

    # --- Init parameters, variables
    if not isinstance(target,int):
        init_nodes_pos = np.vstack((init_nodes_pos,target))
        init_edges_max+= 1
        target = len(init_nodes_pos)-1
    if Mu is None:
        Mu = _PG["Mu"]
    if Lambda is None:
        Lambda = _PG["Lambda"]
    if cycle_Mu is None:
        cycle_Mu = Mu
    if cycle_Lambda is None:
        cycle_Lambda = Lambda
    if n_nodes is None:
        n_nodes = min(16, max(6, len(init_nodes_pos) / 20))

    SquaredX = np.sum(X ** 2, axis=1, keepdims=1)
    part, part_dist = elpigraph.src.core.PartitionData(
        X, init_nodes_pos, 10 ** 6, SquaredX=SquaredX
    )
    clus = (part == source) | (part == target)
    X_fit = np.vstack((init_nodes_pos[source], init_nodes_pos[target], X[clus.flat]))

    # --- fit path
    PG_path = elpigraph.computeElasticPrincipalCurve(
        X_fit, NumNodes=n_nodes, Lambda=Lambda, Mu=Mu, FixNodesAtPoints=[[0], [1]],
    )[0]

    # --- get nodep, edges, create new graph with added loop
    nodep, edges = PG_path["NodePositions"], PG_path["Edges"][0]

    _edges = edges.copy()
    _edges[(edges != 0) & (edges != 1)] += init_edges_max - 1
    _edges[edges == 0] = source
    _edges[edges == 1] = target
    _merged_edges = np.concatenate((init_edges, _edges))
    _merged_nodep = np.concatenate((init_nodes_pos, nodep[2:]))

    if refit_graph:
        cycle_nodes = elpigraph._graph_editing.find_all_cycles(
            nx.Graph(_merged_edges.tolist())
        )

        if len(cycle_nodes) > 0:
            cycle_nodes = flatten(cycle_nodes)

        ElasticMatrix = elpigraph.src.core.MakeUniformElasticMatrix_with_cycle(
            _merged_edges,
            Lambda=Lambda,
            Mu=Mu,
            cycle_Lambda=cycle_Lambda,
            cycle_Mu=cycle_Mu,
            cycle_nodes=cycle_nodes,
        )

        (
            _merged_nodep,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
            X, _merged_nodep, ElasticMatrix, PointWeights=weights, FixNodesAtPoints=[],
        )

    # check intersection
    if _merged_nodep.shape[1] == 2:
        intersect = not (
            MultiLineString(
                [LineString(_merged_nodep[e]) for e in _merged_edges]
            ).is_simple
        )
        if intersect:
            raise ValueError("The created path would intersect existing graph")

    _PG["NodePositions"] = _merged_nodep
    _PG["Edges"] = [_merged_edges]
    _PG["Lambda"] = Lambda
    _PG["Mu"] = Mu
    _PG["cycle_Lambda"] = cycle_Lambda
    _PG["cycle_Mu"] = cycle_Mu
    return _PG


def delPath(
    X,
    PG,
    source,
    target,
    nodes_to_include=None,
    weights=None,
    refit_graph=False,
    Mu=None,
    Lambda=None,
    cycle_Mu=None,
    cycle_Lambda=None,
):
    """
    Delete path between two nodes in the principal graph

    """
    _PG = deepcopy(PG)

    # --- Init parameters, variables
    if Mu is None:
        Mu = _PG["Mu"]
    if Lambda is None:
        Lambda = _PG["Lambda"]
    if cycle_Mu is None:
        cycle_Mu = Mu
    if cycle_Lambda is None:
        cycle_Lambda = Lambda

    # --- get path to remove
    epg_edge = _PG["Edges"][0]
    epg_edge_len = _PG["projection"]["edge_len"].copy()
    del _PG["projection"]

    G = nx.Graph()
    G.add_nodes_from(range(_PG["NodePositions"].shape[0]))
    edges_weighted = list(zip(epg_edge[:, 0], epg_edge[:, 1], epg_edge_len))
    G.add_weighted_edges_from(edges_weighted, weight="len")

    if nodes_to_include is None:
        # nodes on the shortest path
        nodes_sp = nx.shortest_path(G, source=source, target=target, weight="len")
    else:
        assert isinstance(nodes_to_include, list), "`nodes_to_include` must be list"
        # lists of simple paths, in order from shortest to longest
        list_paths = list(
            nx.shortest_simple_paths(G, source=source, target=target, weight="len")
        )
        flag_exist = False
        for p in list_paths:
            if set(nodes_to_include).issubset(p):
                nodes_sp = p
                flag_exist = True
                break
        if not flag_exist:
            print(f"no path that passes {nodes_to_include} exists")

    G.remove_edges_from(np.vstack((nodes_sp[:-1], nodes_sp[1:])).T)
    isolates = list(nx.isolates(G))
    G.remove_nodes_from(isolates)
    Gdel = nx.relabel_nodes(G, dict(zip(G.nodes, np.arange(len(G.nodes)))))

    _PG["Edges"] = [np.array(Gdel.edges)]
    _PG["NodePositions"] = _PG["NodePositions"][
        ~np.isin(range(len(_PG["NodePositions"])), isolates)
    ]

    # --- get nodep, edges, create new graph with added loop
    if refit_graph:
        nodep, edges = _PG["NodePositions"], _PG["Edges"][0]

        cycle_nodes = elpigraph._graph_editing.find_all_cycles(nx.Graph(edges.tolist()))
        if len(cycle_nodes) > 0:
            cycle_nodes = flatten(cycle_nodes)

        ElasticMatrix = elpigraph.src.core.MakeUniformElasticMatrix_with_cycle(
            edges,
            Lambda=Lambda,
            Mu=Mu,
            cycle_Lambda=cycle_Lambda,
            cycle_Mu=cycle_Mu,
            cycle_nodes=cycle_nodes,
        )

        (
            newnodep,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
            X, nodep, ElasticMatrix, PointWeights=weights, FixNodesAtPoints=[]
        )
        _PG["NodePositions"] = newnodep
    return _PG


def refitGraph(
    X,
    PG,
    shift_nodes_pos={},
    PointWeights=None,
    Mu=None,
    Lambda=None,
    cycle_Mu=None,
    cycle_Lambda=None,
):

    init_nodes_pos, init_edges = PG["NodePositions"], PG["Edges"][0]

    # --- Init parameters, variables
    if Mu is None:
        Mu = PG["Mu"]
    if Lambda is None:
        Lambda = PG["Lambda"]
    if cycle_Mu is None:
        cycle_Mu = Mu
    if cycle_Lambda is None:
        cycle_Lambda = Lambda

    # ---Modify node pos values and order (first nodes are fixed)
    for k, v in shift_nodes_pos.items():
        init_nodes_pos[k] = v

    fix_nodes = sorted(list(shift_nodes_pos.keys()), reverse=True)
    fix_order = np.array(fix_nodes+[i for i in range(len(init_nodes_pos)) if i not in fix_nodes])
    fix_nodes_pos = init_nodes_pos[fix_order]
    fix_edges = init_edges.copy()
    for i, ifix in enumerate(fix_order):
        fix_edges[init_edges == ifix] = i

    SquaredX = np.sum(X ** 2, axis=1, keepdims=1)
    part, part_dist = elpigraph.src.core.PartitionData(
        X, fix_nodes_pos, 10 ** 6, SquaredX=SquaredX
    )

    cycle_nodes = find_all_cycles(nx.Graph(fix_edges.tolist()))
    if len(cycle_nodes) > 0:
        cycle_nodes = flatten(cycle_nodes)

    ElasticMatrix = elpigraph.src.core.MakeUniformElasticMatrix_with_cycle(
        fix_edges,
        Lambda=Lambda,
        Mu=Mu,
        cycle_Lambda=cycle_Lambda,
        cycle_Mu=cycle_Mu,
        cycle_nodes=cycle_nodes,
    )

    (
        new_nodes_pos,
        _,
        _,
        _,
        _,
        _,
        _,
    ) = elpigraph.src.core.PrimitiveElasticGraphEmbedment(
        X,
        fix_nodes_pos,
        ElasticMatrix,
        PointWeights=PointWeights,
        FixNodesAtPoints=[[] for i in range(len(fix_nodes))],
    )

    # ---Revert to initial node ordering
    PG["NodePositions"] = new_nodes_pos[np.argsort(fix_order)]




import pytest
import numpy as np
import matplotlib.pyplot as plt
import elpigraph


@pytest.fixture
def data():
    X = np.genfromtxt("./data/tree_data.csv", delimiter=",")
    return X


# test default and non-default parameters
def test_elpi_params(data):
    # Create desired list of inputs for R and Python
    input_data = [data] * 5
    epg_n_nodes = [20, 25, 15, 20, 30]
    epg_lambda = [0.1, 0.02, 0.7, 0.03, 0.08]
    epg_mu = [0.02, 0.07, 0.01, 0.04, 0.06]
    epg_trimmingradius = [float("inf"), 0.7, 0.8, 0.6, 0.5]
    epg_finalenergy = ["Penalized", "Base", "Penalized", "Base", "Base"]
    epg_alpha = [0.01, 0.03, 0.05, 0.08, 0.04]
    epg_beta = [0.03, 0.02, 0.04, 0.07, 0.01]
    epg_mode = [2, 1, 1, 2, 1]
    epg_n_processes = [1, 2, 1, 2, 1]
    epg_collapse_mode = [
        "PointNumber",
        "PointNumber_Extrema",
        "PointNumber_Leaves",
        "EdgesNumber",
        "EdgesLength",
    ]
    epg_collapse_par = [5, 7, 4, 6, 3]
    epg_maxsteps = [float("inf"), 1000, 100, 20, 200]
    # Python uses WeightedCentroid not Weigthed (corrected typo)
    epg_ext_mode = [
        "QuantDists",
        "QuantCentroid",
        "WeightedCentroid",
        "QuantCentroid",
        "WeightedCentroid",
    ]
    r_epg_ext_mode = [
        "QuantDists",
        "QuantCentroid",
        "WeigthedCentroid",
        "QuantCentroid",
        "WeigthedCentroid",
    ]
    epg_ext_par = [0.5, 0.6, 0.8, 0.9, 0.5]
    epg_shift_mode = [
        "NodeDensity",
        "NodePoints",
        "NodeDensity",
        "NodePoints",
        "NodeDensity",
    ]
    epg_shift_radius = [0.05, 0.07, 0.04, 0.08, 0.03]
    epg_shift_max = [5, 7, 4, 8, 6]

    # Results storage Python
    epg_main = []
    epg_obj_collapse = []
    epg_obj_shift = []
    epg_obj_extend = []
    epg_obj_fineTune = []

    for i in range(len(input_data)):

        ############################ Run functions, Python version ###################################

        epg_main.append(
            elpigraph.computeElasticPrincipalTree(
                X=input_data[i],
                NumNodes=epg_n_nodes[i],
                Lambda=epg_lambda[i],
                Mu=epg_mu[i],
                TrimmingRadius=epg_trimmingradius[i],
                FinalEnergy=epg_finalenergy[i],
                alpha=epg_alpha[i],
                beta=epg_beta[i],
                Do_PCA=False,
                CenterData=False,
                n_cores=epg_n_processes[i],
                nReps=1,
                EmbPointProb=1.0,
                Mode=epg_mode[i],
                MaxSteps=epg_maxsteps[i],
            )[0]
        )

        # util functions input
        epg_obj = epg_main[i]
        init_nodes_pos = epg_obj["NodePositions"]
        init_edges = epg_obj["Edges"][0]
        #########################################
        epg_obj_collapse.append(
            elpigraph.CollapseBranches(
                X=input_data[i],
                PG=epg_obj,
                Mode=epg_collapse_mode[i],
                ControlPar=epg_collapse_par[i],
            )
        )

        epg_obj_shift.append(
            elpigraph.ShiftBranching(
                X=input_data[i],
                PG=epg_obj,
                TrimmingRadius=epg_trimmingradius[i],
                SelectionMode=epg_shift_mode[i],
                DensityRadius=epg_shift_radius[i],
                MaxShift=epg_shift_max[i],
            )
        )

        epg_obj_extend.append(
            elpigraph.ExtendLeaves(
                X=input_data[i],
                PG=epg_obj,
                TrimmingRadius=epg_trimmingradius[i],
                Mode=epg_ext_mode[i],
                ControlPar=epg_ext_par[i],
                DoSA_maxiter=100,
            )
        )  # number of iterations for simulated annealing
        epg_obj_fineTune.append(
            elpigraph.fineTuneBR(
                X=input_data[i],
                MaxSteps=epg_maxsteps[i],
                Mode=2,
                NumNodes=epg_n_nodes[i],
                InitNodePositions=init_nodes_pos,
                InitEdges=init_edges,
                Lambda=epg_lambda[i],
                Mu=epg_mu[i],
                TrimmingRadius=epg_trimmingradius[i],
                FinalEnergy=epg_finalenergy[i],
                alpha=epg_alpha[i],
                beta=epg_beta[i],
                Do_PCA=False,
                CenterData=False,
                n_cores=epg_n_processes[i],
                nReps=1,
                ProbPoint=1.0,
            )[0]
        )




import numpy as np
import pandas as pd
import igraph
import copy

# Construct igraph objects -------------------------------------------------------------


def ConstructGraph(PrintGraph):
    """
    Generate an igraph object from an ElPiGraph structure

    PrintGraph A principal graph object

    Return
    -------
    An igraph network


    @examples
    """

    Net = igraph.Graph(n=np.max(PrintGraph["Edges"][0]) + 1, directed=False)
    Net.vs["name"] = list(range(np.max(PrintGraph["Edges"][0]) + 1))
    Net.add_edges(PrintGraph["Edges"][0])

    return Net


# Extract a subpath from the graph ----------------------------------------


def GetSubGraph(Net, Structure, Nodes=None, Circular=True, KeepEnds=True):
    """
    Extract a subgraph with a given topology from a graph

    Net an igraph network object
    Structure a string specifying the structure to return. The following options are
    available:
    \itemize{
     \item 'circle', all the circles of a given length (specified by Nodes) present in the data.
     If Nodes is unspecified the algorithm will look for the largest circle avaialble.
     \item 'branches', all the linear path connecting the branching points
     \item 'branches&bpoints', all the linear path connecting the branching points and all of
     the branching points
     \item 'branching', all the subtree associted with a branching point (i.e., a tree encompassing
    the branching points and the closests branching points and end points)
    \item 'end2end', all linear paths connecting end points (or leaf)
    }
    Circular a boolean indicating whether the circle should contain the initial points at the
    beginning and at the end
    Nodes the number of nodes (for cycle detection)
    KeepEnds boolean, should the end points (overlapping between structures) be included when
    Structure = 'branches' or 'branching'

    @description

    Note that all subgraph are returned only once. So, for example, if A and B are two end leaves of a tree
    and 'end2end' is being used, only the path for A to B or the path from Bt o A will be returned.

    Return
    -------
    a list of nodes defining the structures under consideration




    @examples
    """
    if Structure == "auto":
        print("Structure autodetection is not implemented yet")
        return None

    if Structure == "circle":
        print("WARNING : this mode has not been verified. Confirm with R code")

        if Nodes is None:
            print("Looking for the largest cycle")

            for i in reversed(range(3, Net.vcount())):
                RefNet = igraph.Graph.ring(n=i, directed=False, circular=True)
                if Net.subisomorphic_lad(RefNet):
                    print("A cycle of length", i, "has been found")
                    break

        else:
            i = Nodes

        RefNet = igraph.Graph.Ring(n=i, directed=False, circular=True)

        SubIsoProjList = Net.get_subisomorphisms_vf2(RefNet)
        SubIsoProjList = np.sort(SubIsoProjList, axis=0)
        duplicated = pd.DataFrame(SubIsoProjList).duplicated().to_numpy()
        SubIsoProjList = SubIsoProjList[~duplicated, :]

        # names_SubIsoProjList = ["Circle_" + str(i) for i in range(len(SubIsoProjList))]

        if Circular:
            print(list(map(lambda x: [x, x[0]], SubIsoProjList)))

        else:
            print(SubIsoProjList)

    if Structure == "branches":

        if np.any(np.array(Net.degree()) > 2) & np.any(
            np.array(Net.degree()) == 1
        ):

            # Obtain the branching/end point
            BrPoints = np.where(np.array(Net.degree()) > 2)[0]
            EndPoints = np.where(np.array(Net.degree()) == 1)[0]

            AllPaths = list()

            # Keep track of the interesting nodes
            SelEp = np.union1d(BrPoints, EndPoints)

            for i in range(len(SelEp) - 1):
                AllPaths.extend(
                    Net.get_shortest_paths(
                        SelEp[i],
                        to=SelEp[range(i + 1, len(SelEp))],
                        output="vpath",
                    )
                )

            Valid = np.where(
                np.array(
                    list(
                        map(
                            lambda x: np.sum(np.array(Net.degree(x)) != 2),
                            AllPaths,
                        )
                    )
                )
                == 2
            )[0]

            AllPaths = [AllPaths[i] for i in Valid]

            if not KeepEnds:
                AllPaths = list(
                    map(lambda x: set(x).difference(BrPoints), AllPaths)
                )

            names_AllPaths = ["Branch" + str(i) for i in range(len(AllPaths))]

            CapturedNodes = [item for sublist in AllPaths for item in sublist]
            StillToCapture = set(range(Net.vcount())).difference(CapturedNodes)

            if len(StillToCapture) > 0:
                print(
                    "Unassigned nodes detected. This is due to the presence of"
                    " loops. Additional branching assignement will be"
                    " performed."
                    + "WARNING : case not verified, compare with R version"
                )

                # computing all the distances between the unassigned points and the interesting points
                AllDists = Net.shortest_paths(StillToCapture, SelEp)

                # get the closest interesting point
                EndPoint_1 = np.min(AllDists, axis=1, keepdims=1)

                EndPoint_2 = EndPoint_1.copy()
                EndPoint_2[:, :] = np.nan

                # to get the second we have to avoid passing trough the first one
                for i in range(len(StillToCapture)):
                    tNet = Net.delete_vertices(EndPoint_1[i])
                    PointDists = tNet.shortest_paths(
                        StillToCapture[i],
                        to=set(SelEp).intersection(tNet.vs["name"]),
                    )

                    EndPoint_2[i] = PointDists[np.argmin(PointDists)]

                EndPoints = np.vstack((EndPoint_1, EndPoint_2))
                EndPoints = np.sort(EndPoints, axis=0)

                NewBrEP = EndPoints[:, ~pd.DataFrame(EndPoints.T).duplicated()]
                # NewBrEp = unique_rows = np.unique(original_array, axis=1)
                for i in range(NewBrEP.shape[1]):
                    # for each pair of interesting points
                    # print(i)
                    # Create a temporary network by merging the path with the end points
                    tNet = Net.induced_subgraph(
                        set(
                            StillToCapture[
                                np.all(
                                    np.isin(EndPoints, NewBrEP[:, i]), axis=0
                                )
                            ]
                        ).union(NewBrEP[:, i])
                    )

                    if Net.are_connected(NewBrEP[0, i], NewBrEP[1, i]):
                        tNet.delete_edges(Net.get_eids(NewBrEP[:, i]))

                    tNet = tNet.induced_subgraph(tNet.degree() > 0)

                    # plot(tNet)

                    PotentialEnds = set(NewBrEP[:, i]).intersection(
                        tNet.vs["name"]
                    )

                    if len(PotentialEnds) == 1:
                        # it's a simple loop

                        # get all loops
                        AllLoops = tNet.get_isomorphisms_vf2(
                            igraph.Graph.Ring(
                                tNet.vcount(), directed=False, circular=True
                            )
                        )

                        # select one with the branching point at the beginning
                        Sel = np.where(AllLoops == PotentialEnds)[0][0]

                        # Add a new branch
                        AllPaths.append(AllLoops[Sel])
                        names_AllPaths.append("Branch_" + str(len(AllPaths)))

                    if len(PotentialEnds) == 2:
                        # it's either a line or a line and a loop

                        LinePath = tNet.get_shortest_paths(
                            PotentialEnds[1], PotentialEnds[2]
                        )

                        # Add a new branch
                        AllPaths.append(LinePath)
                        names_AllPaths.append("Branch_" + str(len(AllPaths)))

                        # tNet

                        if len(LinePath) < tNet.vcount():
                            # assuming that the remaining part is a loop do as above
                            AllLoops = tNet.get_subisomorphisms_vf2(
                                tNet,
                                igraph.Graph.Ring(
                                    tNet.vcount() - len(LinePath) + 1,
                                    directed=False,
                                    circular=True,
                                ),
                            )

                            if len(AllLoops) == 0:
                                raise ValueError(
                                    "Unsupported structure. Contact the"
                                    " package maintainer"
                                )

                            # select one with the branching point at the beginning
                            Sel = np.where(np.isin(AllLoops, PotentialEnds))[
                                0
                            ][0]

                            if Sel == np.array([]):
                                raise ValueError(
                                    "Unsupported structure. Contact the"
                                    " package maintainer"
                                )

                            # Add a new branch
                            AllPaths.append(AllLoops[Sel])
                            names_AllPaths.append(
                                "Branch_" + str(len(AllPaths))
                            )

            AllPaths = [i for i in AllPaths if len(i) > 0]

            return AllPaths

        else:
            Structure = "end2end"

    if Structure == "branches&bpoints":

        if np.any(np.array(Net.degree()) > 2) & np.any(
            np.array(Net.degree()) == 1
        ):

            # Obtain the branching/end point
            BrPoints = np.where(np.array(Net.degree()) > 2)[0]
            EndPoints = np.where(np.array(Net.degree()) == 1)[0]

            Allbr = list()
            SelEp = np.union1d(BrPoints, EndPoints)

            for i in BrPoints:

                SelEp = set(SelEp).difference([i])

                for j in SelEp:
                    Path = Net.get_shortest_paths(i, to=j, output="vpath")[0]

                    if not any(
                        np.isin(
                            np.array(Path),
                            np.array(list(set(BrPoints).difference([i, j]))),
                        )
                    ):
                        Allbr.append(Path)

            Allbr = list(
                map(lambda x: list(set(x).difference(BrPoints)), Allbr)
            )

            BaseNameVect = ["Branch" + str(i) for i in range(len(Allbr))]

            BrCount = 0

            for i in BrPoints:
                BrCount = BrCount + 1
                Allbr.append([i])
                BaseNameVect.append(["Branch" + str(BrCount)])

            # names_Allbr = BaseNameVect

            return Allbr

        else:
            Structure == "end2end"

    if Structure == "branching":
        raise ValueError("Not implemented")
    #        BrPoints = np.where(igraph.degree(Net)>2)
    #
    #        Allbr = list()
    #
    #        for i in BrPoints:
    #            Points = i
    #            DONE = False
    #            Terminal_Branching = None
    #            while not DONE:
    #                Nei = unlist(igraph.neighborhood(Net, 1, Points))
    #                Nei = set(Nei).difference(Points)
    #
    #                NeiDeg = igraph.degree(Net, Nei, loops = False)
    #                NewPoints = union(Points, Nei[NeiDeg < 3])
    #
    #                Terminal_Branching = union(Terminal_Branching, Nei[NeiDeg >= 3])
    #
    #                if(len(set(NewPoints).difference(Points)) == 0):
    #                    DONE = True
    #                else:
    #                    Points = NewPoints
    #
    #
    #            if(KeepEnds):
    #                Allbr.append(set(Points).union(Terminal_Branching))
    #
    #            else:
    #                Allbr.append(Points)
    #
    #
    #            names_Allbr = ["Subtree_"+str(i) for i in range(len(Allbr))]
    #
    #        return(Allbr)

    if Structure == "end2end":

        EndPoints = np.where(np.array(Net.degree()) == 1)[0]

        Allbr = list()

        for i in range(len(EndPoints) - 1):
            for j in range(i + 1, len(EndPoints)):
                Path = Net.get_shortest_paths(
                    EndPoints[i], to=EndPoints[j], output="vpath"
                )
                if len(Path) > 0:
                    Allbr.append(Path)

        # names_Allbr = ["Path_" + str(i) for i in range(len(Allbr))]

        return Allbr


def GetBranches(Net, StartingPoints=None):
    """
    Return a dict summarizing the branching structure

    Net an igraph network
    StartingPoint the starting points

    Return
    -------
    a data frame with three columns:
    \itemize{
     \item VName contains the vertx names
     \item Branch contains the id of the branch (numbering start from the branch
    containing the starting point) or 0 if a branching point
     \item BrPoints the branchhing point id, or 0 if not a branching point
    }



    @examples
    """
    # Net = ConstructGraph(ReturnList[0])
    StartingPoint = None

    if StartingPoint is None:
        EndPoints = np.where(np.array(Net.degree()) == 1)[0]
        StartingPoint = np.random.choice(EndPoints, 1)

    if Net.degree(StartingPoint)[0] != 1:
        raise ValueError("Invalid starting point")

    Vertices = Net.vs["name"]

    Branches = np.zeros(len(Vertices))
    DiffPoints = np.zeros(len(Vertices))

    # names_Branches = Vertices
    # names_DiffPoints = Vertices

    tNet = copy.deepcopy(Net)
    CurrentEdges = StartingPoint
    Branches[StartingPoint] = 1
    NewEdges = set()

    while len(tNet.vs.indices) > 0:

        for i in range(len(CurrentEdges)):
            AllNei = [
                tNet.vs["name"][k]
                for k in tNet.neighborhood(
                    vertices=tNet.vs["name"].index(CurrentEdges[i]), order=1
                )
            ]
            AllNei = list(set(AllNei).difference([CurrentEdges[i]]))

            NewEdges = NewEdges.union(AllNei)
            if len(AllNei) > 0:
                for j in range(len(AllNei)):
                    if tNet.degree(tNet.vs["name"].index(AllNei[j])) > 2:
                        DiffPoints[AllNei[j]] = np.max(DiffPoints) + 1

                    else:
                        if DiffPoints[CurrentEdges[i]] > 0:
                            Branches[AllNei[j]] = np.max(Branches) + 1

                        else:
                            Branches[AllNei[j]] = Branches[CurrentEdges[i]]
            #                 print(DiffPoints)
            tNet.delete_vertices(tNet.vs["name"].index(CurrentEdges[i]))

        CurrentEdges = list(NewEdges)
        CurrentEdges.sort()
        NewEdges = set()

    return dict(VName=Vertices, Branch=Branches, BrPoints=DiffPoints)



import numpy as np


def InitializeWalkers(Number=1, Dimensions=3, MeanShift=1, SdShift=0.3):
    """
    Initialize a group of 'walkers'

    Number integer, the number of walkers to be initialised
    Dimensions integer, the number of dimensions of the walkers
    MeanShift numeric, the mean shift in each dimension for the walker
    SdShift positive numeric, the standard deviation in the shift in each dimension for the walker

    Return
    -------
    A list with the walkers


    @examples

    Walkers <- InitializeWalkers(Dimensions = 1000)

    """
    RetList = list()

    for i in range(Number):
        RetList.append(
            dict(
                nDim=Dimensions,
                Positions=np.zeros(Dimensions),
                DiffusionShift=np.random.choice([-1, 1], 10)
                * np.random.normal(MeanShift, SdShift, Dimensions),
                Name=i,
                Age=0,
            )
        )

    return RetList


def GrowPath(
    Walkers, StepSD=0.1, nSteps=100, BranchProb=0.01, MinAgeBr=50, BrDim=5
):
    """
    Grow a branching path using walkers

    Walkers list, a list of walker as returned from the InizializeWalkers function
    StepSize positive numeric, the standard deviation associated with the movement
    in each direction for each step
    nSteps integer, the number of steps
    BranchProb numeric between 0 and 1, the probability per walker of branching at each step
    MinAgeBr integer, the minimal number of steps before a newly introduced walker will start branhing
    BrDim integer, the number of dimensions affected during branching

    @return


    @examples

    Walkers <- InizializeWalkers(nDim = 1000)

    Data <- GrowPath(Walkers = Walkers, StepSize = 50, nSteps = 2000, BranchProb = .0015, MinAgeBr = 75, BrDim = 15)

    """
    Trace = np.zeros((1, len(Walkers[0]["Pos"])))
    Branch = []
    TimeVect = []
    TimeVal = -1

    while len(Trace) < nSteps:
        nWalkers = len(Walkers)
        TimeVal = TimeVal + 1
        for j in range(nWalkers):
            # Create a new point by diffusion
            Walkers[j]["Pos"] = (
                Walkers[j]["Pos"]
                + Walkers[j]["Dif"]
                + np.random.normal(size=Walkers[j]["nDim"], scale=StepSD)
            )
            Trace = np.vstack(Trace, Walkers[j]["Pos"])
            # Keep track of the time
            TimeVect.append(TimeVal)
            # In crese the age of the walker
            Walkers[j]["Age"] = Walkers[j]["Age"] + 1
            # Keep Track of the originating population
            Branch.append(Walkers[j]["Name"])
            if (np.random.uniform() < BranchProb) & (
                Walkers[j]["Age"] >= MinAgeBr
            ):
                # Create a new branch
                BranchDim = (
                    np.random.uniform(size=Walkers[j]["nDim"])
                    < BrDim / Walkers[j]["nDim"]
                )
                if np.any(BranchDim) & (~np.all(BranchDim)):
                    print(
                        "Time=",
                        len(Trace),
                        " Branching in ",
                        sum(BranchDim),
                        " dimensions \n",
                    )

                    Walkers[j]["Age"] = 0
                    Walkers[j]["Name"] = max([i["Name"] for i in Walkers]) + 1

                    NewBuild = Walkers[j]
                    NewBuild["Age"] = 0
                    NewBuild["Dif"][BranchDim] = -NewBuild["Dif"][BranchDim]
                    NewBuild["Name"] = Walkers[j]["Name"] + 1
                    Walkers.append(NewBuild)

    return dict(
        UpdatedWalker=Walkers, Trace=Trace, Branch=Branch, Time=TimeVect
    )



import numpy as np
import pandas as pd
import networkx as nx
import scipy
import networkx as nx
from .src.core import PartitionData
from .src.reporting import project_point_onto_graph
from sklearn.neighbors import NearestNeighbors, KernelDensity


def getProjection(X, PG):
    """Compute graph projection from principal graph dict
    (result stored in PG dict under 'projection' key)
    """

    G = nx.Graph()
    G.add_edges_from(PG["Edges"][0].tolist(), weight=1)
    mat_conn = nx.to_scipy_sparse_matrix(
        G, nodelist=np.arange(len(PG["NodePositions"])), weight="weight"
    )

    # partition points
    node_id, node_dist = PartitionData(
        X=X,
        NodePositions=PG["NodePositions"],
        MaxBlockSize=len(PG["NodePositions"]) ** 4,
        SquaredX=np.sum(X ** 2, axis=1, keepdims=1),
    )
    # project points onto edges
    dict_proj = project_point_onto_graph(
        X=X,
        NodePositions=PG["NodePositions"],
        Edges=PG["Edges"][0],
        Partition=node_id,
    )

    PG["projection"] = {}
    PG["projection"]["node_id"] = node_id.flatten()
    PG["projection"]["node_dist"] = node_dist
    PG["projection"]["edge_id"] = dict_proj["EdgeID"].astype(int)
    PG["projection"]["edge_loc"] = dict_proj["ProjectionValues"]
    PG["projection"]["X_projected"] = dict_proj["X_projected"]
    PG["projection"]["conn"] = mat_conn
    PG["projection"]["edge_len"] = dict_proj["EdgeLen"]
    PG["projection"]["MSEP"] = dict_proj["MSEP"]


def getPseudotime(
    X, PG, source, target=None, nodes_to_include=None, project=True
):
    """Compute pseudotime given
    source: int
        source node
    target: int
        optional target node
    nodes_to_include: list
        optional nodes to include in the path
        (useful for complex topologies with loops,
        where multiple paths are possible between 2 nodes)
    project: bool
        if False, will save computation time by using the projection already stored in PG dict (computed using elpigraph.utils.getProjection)
    """
    if project is True:
        getProjection(X, PG)
    elif project is False and not ("projection" in PG.keys()):
        raise ValueError(
            "key 'projection not found in PG. To use a precomputed projection"
            " run elpigraph.utils.getProjection"
        )

    epg_edge = PG["Edges"][0]
    epg_edge_len = PG["projection"]["edge_len"]
    G = nx.Graph()
    edges_weighted = list(zip(epg_edge[:, 0], epg_edge[:, 1], epg_edge_len))
    G.add_weighted_edges_from(edges_weighted, weight="len")
    if target is not None:
        if nodes_to_include is None:
            # nodes on the shortest path
            nodes_sp = nx.shortest_path(
                G, source=source, target=target, weight="len"
            )
        else:
            assert isinstance(
                nodes_to_include, list
            ), "`nodes_to_include` must be list"
            # lists of simple paths, in order from shortest to longest
            list_paths = list(
                nx.shortest_simple_paths(
                    G, source=source, target=target, weight="len"
                )
            )
            flag_exist = False
            for p in list_paths:
                if set(nodes_to_include).issubset(p):
                    nodes_sp = p
                    flag_exist = True
                    break
            if not flag_exist:
                return f"no path that passes {nodes_to_include} exists"
    else:
        nodes_sp = [source] + [v for u, v in nx.bfs_edges(G, source)]
    G_sp = G.subgraph(nodes_sp).copy()
    index_nodes = {
        x: nodes_sp.index(x) if x in nodes_sp else G.number_of_nodes()
        for x in G.nodes
    }

    if target is None:
        dict_dist_to_source = nx.shortest_path_length(
            G_sp, source=source, weight="len"
        )
    else:
        dict_dist_to_source = dict(
            zip(
                nodes_sp,
                np.cumsum(
                    np.array(
                        [0.0]
                        + [
                            G.get_edge_data(nodes_sp[i], nodes_sp[i + 1])[
                                "len"
                            ]
                            for i in range(len(nodes_sp) - 1)
                        ]
                    )
                ),
            )
        )

    cells = np.isin(PG["projection"]["node_id"], nodes_sp)
    id_edges_cell = PG["projection"]["edge_id"][cells].tolist()
    edges_cell = PG["Edges"][0][id_edges_cell, :]
    len_edges_cell = PG["projection"]["edge_len"][id_edges_cell]

    # proportion on the edge
    prop_edge = np.clip(PG["projection"]["edge_loc"][cells], a_min=0, a_max=1)

    dist_to_source = []
    for i in np.arange(edges_cell.shape[0]):
        if index_nodes[edges_cell[i, 0]] > index_nodes[edges_cell[i, 1]]:
            dist_to_source.append(dict_dist_to_source[edges_cell[i, 1]])
            prop_edge[i] = 1 - prop_edge[i]
        else:
            dist_to_source.append(dict_dist_to_source[edges_cell[i, 0]])
    dist_to_source = np.array(dist_to_source)
    dist_on_edge = len_edges_cell * prop_edge
    dist = dist_to_source + dist_on_edge

    PG["pseudotime"] = np.repeat(np.nan, len(X))
    PG["pseudotime"][cells] = dist
    PG["pseudotime_params"] = {
        "source": source,
        "target": target,
        "nodes_to_include": nodes_to_include,
    }


#### supervised knn


def _longform_knn_to_sparse(dis, idx):
    row_ind = np.tile(np.arange(len(idx))[:, None], idx.shape[1])
    col_ind = idx
    return scipy.sparse.csr_matrix((dis.flat, (row_ind.flat, col_ind.flat)))


def getWeights(
    X, bandwidth=1, griddelta=100, exponent=1, method="sklearn", **kwargs
):
    """Get point weights as the inverse density of data
    X: np.array, (n_sample x n_dims)
    bandwidth: sklearn KernelDensity bandwidth if method == 'sklearn'
    griddelta: FFTKDE grid step size if method =='fft'
    exponent: density values are raised to the power of exponent
    """
    if method == "sklearn":
        kde = KernelDensity(
            kernel="gaussian", bandwidth=bandwidth, **kwargs
        ).fit(X)
        scores = kde.score_samples(X)
        scores = np.exp(scores)[:, None]

    elif method == "fft":
        import KDEpy

        kde = KDEpy.FFTKDE(**kwargs).fit(X)
        x, y = kde.evaluate(griddelta)
        scores = scipy.interpolate.griddata(x, y, X)

    p = 1 / (scores ** exponent)
    p /= np.sum(p)
    return p


def ordinal_neighbors_stagewise_longform(
    X, stages_labels, stages=None, k=15, radius=None, m="cosine"
):
    """Supervised (ordinal) nearest-neighbor search.
    Stages is an ordered list of stages labels (low to high). If None, taken as np.unique(stages_labels)"""

    if stages is None:
        stages = np.unique(stages_labels)

    knn_distances = np.zeros((len(X), 3 * k))
    knn_idx = np.zeros((len(X), 3 * k), dtype=int)

    nn_stage = {}
    for s in stages:
        nn_stage[s] = NearestNeighbors(n_neighbors=k, metric=m).fit(
            X[stages_labels.values == s, :]
        )

    s = []
    t = []
    w = []
    for i in range(len(stages) - 1):
        dis, ind = nn_stage[stages[i]].kneighbors(
            X[stages_labels.values == stages[i + 1], :], k
        )
        knn_distances[stages_labels.values == stages[i + 1], :k] = dis
        knn_idx[stages_labels.values == stages[i + 1], :k] = np.where(
            stages_labels.values == stages[i]
        )[0][ind]

    for i in range(1, len(stages)):
        dis, ind = nn_stage[stages[i]].kneighbors(
            X[stages_labels.values == stages[i - 1], :], k
        )
        knn_distances[stages_labels.values == stages[i - 1], k : 2 * k] = dis
        knn_idx[stages_labels.values == stages[i - 1], k : 2 * k] = np.where(
            stages_labels.values == stages[i]
        )[0][ind]

    for i in range(len(stages)):
        if i == 0:
            dis, ind = nn_stage[stages[i]].kneighbors(
                X[stages_labels.values == stages[i], :], 2 * k + 1
            )
            dis, ind = dis[:, 1:], ind[:, 1:]
            knn_distances[
                np.argwhere(stages_labels.values == stages[i]),
                list(range(k)) + list(range(2 * k, 3 * k)),
            ] = dis
            knn_idx[
                np.argwhere(stages_labels.values == stages[i]),
                list(range(k)) + list(range(2 * k, 3 * k)),
            ] = np.where(stages_labels.values == stages[i])[0][ind]

        elif i == (len(stages) - 1):
            dis, ind = nn_stage[stages[i]].kneighbors(
                X[stages_labels.values == stages[i], :], 2 * k + 1
            )
            dis, ind = dis[:, 1:], ind[:, 1:]
            knn_distances[stages_labels.values == stages[i], k:] = dis
            knn_idx[stages_labels.values == stages[i], k:] = np.where(
                stages_labels.values == stages[i]
            )[0][ind]

        else:
            dis, ind = nn_stage[stages[i]].kneighbors(
                X[stages_labels.values == stages[i], :], k + 1
            )
            dis, ind = dis[:, 1:], ind[:, 1:]
            knn_distances[stages_labels.values == stages[i], 2 * k :] = dis
            knn_idx[stages_labels.values == stages[i], 2 * k :] = np.where(
                stages_labels.values == stages[i]
            )[0][ind]

    _sort = np.argsort(knn_distances, axis=1)
    knn_distances = knn_distances[
        np.arange(len(knn_distances))[:, None], _sort
    ]
    knn_idx = knn_idx[np.arange(len(knn_distances))[:, None], _sort]

    return knn_distances, knn_idx


def ordinal_neighbors_longform(
    X, stages_labels, stages=None, k=15, m="cosine"
):
    """Supervised (ordinal) nearest-neighbor search.
    Stages is an ordered list of stages labels (low to high). If None, taken as np.unique(stages_labels)"""
    if stages is None:
        stages = np.unique(stages_labels)

    knn_distances = np.zeros((len(X), k))
    knn_idx = np.zeros((len(X), k), dtype=int)
    for i in range(len(stages)):

        sel_points = (
            (stages_labels.values == stages[i])
            | (stages_labels.values == stages[max(0, i - 1)])
            | (stages_labels.values == stages[min(i + 1, len(stages) - 1)])
        )

        stage = stages_labels.values == stages[i]
        dis, ind = (
            NearestNeighbors(n_neighbors=k, metric=m)
            .fit(X[sel_points, :])
            .kneighbors(X[stage, :])
        )

        knn_distances[stage, :] = dis
        knn_idx[stage, :] = np.where(sel_points)[0][ind]
    return knn_distances, knn_idx


def supervised_knn(
    X,
    stages_labels,
    stages=None,
    n_neighbors=15,
    n_natural=0,
    m="cosine",
    method="force",
    return_sparse=False,
):
    """Supervised (ordinal) nearest-neighbor search.
    Stages is an ordered list of stages labels (low to high). If None, taken as np.unique(stages_labels)

    Parameters
    ----------
    method : str (default='force')
        if 'force', searches for each point at stage[i] n_neighbors nearest_neighbors, forcing:
            - n_neighbors/3 to be from stage[i-1]
            - n_neighbors/3 to be from stage[i]
            - n_neighbors/3 to be from stage[i+1]
            For stage[0] and stage[-1], 2*n_neighbors/3 are taken from stage[i]

        if 'guide', searches for each point at stage[i] n_neighbors nearest_neighbors
            from points in {stage[i-1], stage[i], stage[i+1]}, without constraints on proportions
    """

    if n_neighbors % 3 != 0:
        raise ValueError("Please provide n_neighbors divisible by 3")
    if stages is None:
        stages = np.unique(stages_labels)

    dis, idx = (
        NearestNeighbors(n_neighbors=n_neighbors, metric=m, n_jobs=8)
        .fit(X)
        .kneighbors()
    )

    if method == "guide":
        knn_distances, knn_idx = ordinal_neighbors_longform(
            X, stages_labels, stages=stages, k=n_neighbors, m=m
        )
    if method == "force":
        knn_distances, knn_idx = ordinal_neighbors_stagewise_longform(
            X, stages_labels, stages=stages, k=n_neighbors // 3, m=m
        )

    # ---mix natural nn with ordinal nn
    merged_idx = np.zeros((len(X), n_neighbors), dtype=np.int32)
    merged_dists = np.zeros((len(X), n_neighbors))
    for i in range(len(X)):
        merged_idx[i][:n_natural] = idx[i][:n_natural]
        merged_idx[i][n_natural:] = np.setdiff1d(
            knn_idx[i], idx[i][:n_natural], assume_unique=True
        )[: n_neighbors - n_natural]

        merged_dists[i][:n_natural] = dis[i][:n_natural]
        merged_dists[i][n_natural:] = knn_distances[i][
            np.isin(
                knn_idx[i],
                np.setdiff1d(
                    knn_idx[i], idx[i][:n_natural], assume_unique=True
                )[: n_neighbors - n_natural],
            )
        ]

    if return_sparse:
        return _longform_knn_to_sparse(merged_dists, merged_idx)
    else:
        return merged_dists, merged_idx


def geodesic_pseudotime(X, k, root, g=None):
    """pseudotime as graph distance from root point"""
    if g is None:
        nn = NearestNeighbors(n_neighbors=k, n_jobs=8).fit(X)
        g = nx.convert_matrix.from_scipy_sparse_matrix(
            nn.kneighbors_graph(mode="distance")
        )
    else:
        g = nx.convert_matrix.from_scipy_sparse_matrix(g)
    if len(list(nx.connected_components(g))) > 1:
        raise ValueError(
            f"detected more than 1 components with k={k} neighbors. Please"
            " increase k"
        )
    lengths = nx.single_source_dijkstra_path_length(g, root)
    pseudotime = np.array(pd.Series(lengths).sort_index())
    return pseudotime


def proj2embedding(X_elpi,X_embed,NodePositions):
    part,dis = PartitionData(X_elpi,NodePositions,50000,np.sum(X_elpi**2,axis=1,keepdims=1))
    # Hard assigment
    R = scipy.sparse.csr_matrix(
        (np.ones(len(X_elpi)), (range(len(X_elpi)), part.flat)), (len(X_elpi),len(NodePositions))
    ).A
    proj = (np.dot(X_embed.T, R) / R.sum(axis=0)).T
    return proj


def _propagate_labels(Xs,Xt,ys,flavor,n_neighbors,reg_e=0.01,reg_m=0.08):
    if 'ot' in flavor:
        try: import ot
        except: raise ImportError('Install Python OT to use this function'+
            '(pip install POT or conda install -c conda-forge pot)')
            
    if flavor=='ot':
        coupling = ot.sinkhorn([], [], ot.da.cost_normalization(ot.dist(Xs,Xt),'max'),reg=reg_e,numItermax=1000000)
    elif flavor=='ot_unbalanced':
        coupling = ot.sinkhorn_unbalanced([], [], ot.da.cost_normalization(ot.dist(Xs,Xt),'max'),reg_e,reg_m,numItermax=1000000)
    elif flavor=='ot_equal':
        ws=np.ones(len(ys))
        ws[ys==0]/=sum(ys==0)
        ws[ys==1]/=sum(ys==1)
        ws/=sum(ws)
        coupling = ot.sinkhorn(ws, [], ot.da.cost_normalization(ot.dist(Xs,Xt),'max'), reg=reg_e,numItermax=1000000)
    elif flavor=='knn':
        coupling = NearestNeighbors(n_neighbors=n_neighbors).fit(Xs).kneighbors_graph(Xt).toarray().T
    else: 
        raise ValueError('method must be ot, ot_unbalanced, ot_equal, knn')
        
    labels=[]
    for j,coup in enumerate(coupling.T):
        nz=coup>0
        labels.append(np.bincount(ys[nz],weights=coup[nz].flat).argmax())
        #labels.append(ys[coup.argmax()])

    return labels

def early_groups(X, PG, branch_nodes, source, target, nodes_to_include = None,
                 flavor='ot_unbalanced', n_windows = 20, n_neighbors=20, ot_reg_e=0.01,ot_reg_m=0.001):

    getPseudotime(X, PG, source, target=target, nodes_to_include=nodes_to_include, project=True)
    ix=~np.isnan(PG['pseudotime'])
    
    y = -np.ones(len(X),dtype=int)

    bnodes_ix = np.zeros(len(X),dtype=bool)
    for bn in branch_nodes:
        nix = PG['projection']['node_id']==bn
        bnodes_ix[nix] = True
        y[nix] = bn
        
    Xpath, ypath = X[ix].copy(), y[ix].copy() # points from path 
    Xbranch, ybranch = X[bnodes_ix].copy(), y[bnodes_ix].copy() # points from branch
    
    #propagate labels
    bpseudotime = PG['pseudotime'][ix]
    count, bins = np.histogram(bpseudotime, bins=n_windows)
    clusters = np.digitize(bpseudotime, bins[1:], right=True)

    #init 
    cnext = clusters==clusters.max()
    ypath[cnext] = _propagate_labels(Xs=Xbranch,Xt=Xpath[cnext],ys=ybranch,flavor=flavor,n_neighbors=n_neighbors)

    for c in np.arange(1,n_windows)[::-1]:
        cprev = clusters==c
        cnext = clusters==(c-1)
        ypath[cnext] = _propagate_labels(Xs=Xpath[cprev],Xt=Xpath[cnext],ys=ypath[cprev],
                                    flavor=flavor,n_neighbors=n_neighbors,reg_e=ot_reg_e,reg_m=ot_reg_m)
        
    s = '-'.join(str(x) for x in branch_nodes)
    
    y[ix] = ypath
    PG[f'early_groups_{source}->{s}']=y.astype(str)
    y[ix] = clusters
    PG[f'early_groups_{source}->{s}_clusters']=y.astype(str)


from ._AlterStructure import ExtendLeaves, CollapseBranches, ShiftBranching
from ._BaseElPiWrapper import computeElasticPrincipalGraphWithGrammars
from ._EMAdjustment import AdjustByConstant
from ._graph_editing import findPaths, addPath, delPath, refitGraph
from ._topologies import (
    computeElasticPrincipalCircle,
    computeElasticPrincipalTree,
    computeElasticPrincipalCurve,
    fineTuneBR,
    GrowLeaves,
    generateInitialConfiguration,
)
from . import src
from . import plot
from . import utils
from ._version import __version__
